var _time = require("./time.js");
var _error = require("./error.js");
var _check = require("./check.js");
const debounce = (func, wait = 250, isLeadingEdge = !1) => (isLeadingEdge ? debounceL : debounceT)(func, wait);
exports.debounce = debounce;
const debounceT = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    clearTimeout(timeoutToken);
    timeoutToken = setTimeout((() => {
      timeoutToken = null;
      func.apply(null, args);
    }), wait);
  };
};
exports.debounceT = debounceT;
const debounceL = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    const isCallNow = null === timeoutToken;
    clearTimeout(timeoutToken);
    timeoutToken = setTimeout((() => {
      timeoutToken = null;
    }), wait);
    isCallNow && func.apply(null, args);
  };
};
exports.debounceL = debounceL;
exports.throttle = (func, wait = 250, isLeadingEdge = !1) => (isLeadingEdge ? throttleL : throttleTE)(func, wait);
const throttleTE = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    timeoutToken || (timeoutToken = (0, _time.setWeakTimeout)((() => {
      timeoutToken = null;
      func.apply(null, args);
    }), wait));
  };
};
exports.throttleT = (func, wait = 250) => {
  let timeoutToken = null;
  let lastArgs = null;
  return (...args) => {
    lastArgs = args;
    timeoutToken || (timeoutToken = (0, _time.setWeakTimeout)((() => {
      timeoutToken = null;
      func.apply(null, lastArgs);
      lastArgs = null;
    }), wait));
  };
};
const throttleL = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    if (timeoutToken) return;
    const isCallNow = null === timeoutToken;
    timeoutToken = (0, _time.setWeakTimeout)((() => {
      timeoutToken = null;
    }), wait);
    isCallNow && func.apply(null, args);
  };
};
exports.throttleL = throttleL;
exports.once = func => {
  let isCalled = !1;
  return (...args) => {
    if (!0 !== isCalled) {
      isCalled = !0;
      return func.apply(null, args);
    }
  };
};
exports.lossyAsync = (asyncFunc, onError = _error.rethrowError) => {
  let runningPromise;
  const onResolve = () => {
    runningPromise = void 0;
  };
  const onReject = error => {
    runningPromise = void 0;
    onError(error);
  };
  return {
    trigger: (...args) => {
      if (!runningPromise) try {
        const result = asyncFunc.apply(null, args);
        (0, _check.isPromiseAlike)(result) ? runningPromise = result.then(onResolve, onReject) : onResolve();
      } catch (error) {
        onReject(error);
      }
    },
    getRunningPromise: () => runningPromise
  };
};
exports.loneAsync = asyncFunc => {
  let runningPromise;
  const onEnd = () => {
    runningPromise = void 0;
  };
  return (...args) => {
    if (runningPromise) return runningPromise;
    const result = asyncFunc.apply(null, args);
    if ((0, _check.isPromiseAlike)(result)) {
      result.then(onEnd, onEnd);
      runningPromise = result;
    }
    return result;
  };
};
exports.withCache = func => {
  let cache;
  return () => {
    void 0 === cache && (cache = func());
    return cache;
  };
};
exports.withCacheAsync = asyncFunc => {
  let cache;
  return async () => {
    void 0 === cache && (cache = await asyncFunc());
    return cache;
  };
};
exports.withDelayArgvQueue = (func, delayWrapper = debounce, ...args) => {
  let argvQueue = [];
  const delayFunc = delayWrapper((() => {
    const currentArgvQueue = argvQueue;
    argvQueue = [];
    func(currentArgvQueue);
  }), ...args);
  return (...argv) => {
    argvQueue.push(argv);
    delayFunc();
  };
};
exports.withRepeat = (func, count = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    func(looped, count);
    looped++;
  }
};
exports.withRepeatAsync = async (func, count = 0, wait = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    const startTime = (0, _time.clock)();
    await func(looped, count);
    looped++;
    const remainingTime = wait - ((0, _time.clock)() - startTime);
    remainingTime > 0 && await (0, _time.setTimeoutAsync)(remainingTime);
  }
};
exports.withRetry = (func, maxRetry = 1 / 0) => {
  let failed = 0;
  for (;;) try {
    return func(failed, maxRetry);
  } catch (error) {
    failed++;
    if (maxRetry < failed) throw error;
  }
};
exports.withRetryAsync = async (func, maxRetry = 1 / 0, wait = 0) => {
  let failed = 0;
  for (;;) {
    const startTime = (0, _time.clock)();
    try {
      return await func(failed, maxRetry);
    } catch (error) {
      failed++;
      if (maxRetry < failed) throw error;
      const remainingTime = wait - ((0, _time.clock)() - startTime);
      remainingTime > 0 && await (0, _time.setTimeoutAsync)(remainingTime);
    }
  }
};
exports.withTimeoutAsync = (func, timeout, message = "") => withTimeoutPromise(func(), timeout, message);
const withTimeoutPromise = (promise, timeout, message = "") => {
  if (!(timeout > 0)) return promise;
  let timeoutToken = null;
  return Promise.race([ promise, new Promise(((resolve, reject) => {
    timeoutToken = (0, _time.setWeakTimeout)((() => reject(DUMMY_ERROR)), timeout);
  })) ]).then((result => {
    clearTimeout(timeoutToken);
    return result;
  }), (error => {
    clearTimeout(timeoutToken);
    throw error !== DUMMY_ERROR ? error : new Error(`timeout after: ${timeout}${message ? `, ${message}` : ""}`);
  }));
};
exports.withTimeoutPromise = withTimeoutPromise;
const DUMMY_ERROR = {};
const createInsideOutPromise = () => {
  let promiseResolve, promiseReject;
  return {
    promise: new Promise(((resolve, reject) => {
      promiseResolve = resolve;
      promiseReject = reject;
    })),
    resolve: promiseResolve,
    reject: promiseReject
  };
};
exports.createInsideOutPromise = createInsideOutPromise;
const runAsPromise = func => {
  try {
    const result = func();
    return (0, _check.isPromiseAlike)(result) ? result : Promise.resolve(result);
  } catch (error) {
    return Promise.reject(error);
  }
};
exports.runAsPromise = runAsPromise;
exports.runAsyncByLane = (laneSize, asyncFuncList = []) => {
  if (0 === (laneSize = Math.min(laneSize, asyncFuncList.length))) return Promise.resolve();
  const afList = [ ...asyncFuncList ];
  const iop = createInsideOutPromise();
  let isAbort = !1;
  let activeTask = 0;
  const onResolve = () => {
    if (!isAbort) {
      activeTask--;
      afList.length || activeTask ? runMore() : iop.resolve();
    }
  };
  const onReject = error => {
    isAbort = !0;
    iop.reject(error);
  };
  const runMore = () => {
    if (!isAbort) for (;activeTask < laneSize && afList.length; ) {
      activeTask++;
      runAsPromise(afList.shift()).then(onResolve, onReject);
    }
  };
  runMore();
  return iop.promise;
};