var _compare = require("../compare.js");
const REGEXP_SEMVER = /^(\d+)\.(\d+)\.(\d+)(-[0-9A-Za-z-.]+)?$/;
const parseSemVer = versionString => {
  let [, major, minor, patch, label = ""] = REGEXP_SEMVER.exec(versionString) || [];
  major = parseInt(major);
  minor = parseInt(minor);
  patch = parseInt(patch);
  if (isNaN(major) || isNaN(minor) || isNaN(patch)) throw new Error(`invalid version: ${versionString}`);
  return {
    major,
    minor,
    patch,
    label
  };
};
exports.parseSemVer = parseSemVer;
exports.packSemVer = ({major, minor, patch, label = ""}) => `${major}.${minor}.${patch}${label}`;
exports.compareSemVer = (stringA, stringB) => {
  const a = parseSemVer(stringA);
  const b = parseSemVer(stringB);
  return a.major - b.major || a.minor - b.minor || a.patch - b.patch || compareSemVerLabel(a.label, b.label);
};
const compareSemVerLabel = (a, b) => a === b ? 0 : "" === a ? 1 : "" === b ? -1 : compareSemVerLabelIdentifier(a, b);
const compareSemVerLabelIdentifier = (a, b) => {
  const aList = a.split(".");
  const bList = b.split(".");
  let index = 0;
  for (;;) {
    const aI = aList[index];
    const bI = bList[index];
    if (aI === bI) {
      index++;
      continue;
    }
    if (void 0 === aI) return -1;
    if (void 0 === bI) return 1;
    const aAllNumber = REGEXP_ALL_NUMBER.test(aI);
    const bAllNumber = REGEXP_ALL_NUMBER.test(bI);
    if (aAllNumber && bAllNumber) return parseInt(aI) - parseInt(bI);
    if (!aAllNumber && bAllNumber) return 1;
    if (aAllNumber && !bAllNumber) return -1;
    if (!aAllNumber && !bAllNumber) return (0, _compare.compareString)(aI, bI);
  }
};
const REGEXP_ALL_NUMBER = /^\d+$/;
exports.versionBumpByGitBranch = (version, {gitBranch, getIsMajorBranch = gitBranch => [ "master", "main", "major" ].includes(gitBranch), isMajorBranch = getIsMajorBranch(gitBranch)}) => {
  const {major, minor, patch, label} = parseSemVer(version);
  if (isMajorBranch) return `${major}.${minor}.${label ? patch : parseInt(patch) + 1}`;
  {
    const labelGitBranch = gitBranch.replace(/[\W_]/g, "");
    return versionBumpToIdentifier(version, {
      identifier: labelGitBranch
    });
  }
};
const versionBumpToIdentifier = (version, {identifier = "TEST"}) => {
  let {major, minor, patch, label} = parseSemVer(version);
  if (label.startsWith(`-${identifier}.`)) {
    const extraString = label.slice(`-${identifier}.`.length);
    if (/^\d+$/.test(extraString)) return versionBumpLastNumber(version);
    const extraStringFirstIdentifier = extraString.split(".")[0];
    if (/^\d+$/.test(extraStringFirstIdentifier)) return `${major}.${minor}.${patch}-${identifier}.${parseInt(extraStringFirstIdentifier) + 1}`;
    label = "";
  }
  return `${major}.${minor}.${label ? patch : patch + 1}-${identifier}.0`;
};
exports.versionBumpToIdentifier = versionBumpToIdentifier;
const versionBumpLastNumber = version => {
  parseSemVer(version);
  if (!REGEXP_LAST_NUMBER.test(version)) throw new Error(`[versionBumpLastNumber] no number to bump in version: ${version}`);
  return version.replace(REGEXP_LAST_NUMBER, ((match, $1, $2) => `${parseInt($1) + 1}${$2}`));
};
exports.versionBumpLastNumber = versionBumpLastNumber;
const REGEXP_LAST_NUMBER = /(\d+)(\D*)$/;
exports.versionBumpToLocal = version => {
  const {major, minor, patch, label} = parseSemVer(version);
  return label ? REGEXP_LABEL_LOCAL.test(label) ? versionBumpLastNumber(version) : `${major}.${minor}.${patch}${label}.local.0` : `${major}.${minor}.${patch}-local.0`;
};
const REGEXP_LABEL_LOCAL = /^-(?:.*\.)?local\.\d+$/;
exports.isVersionSpecComplex = versionSpec => versionSpec.includes(" ") || versionSpec.includes(":") || versionSpec.includes("/");