var _tryRequire = require("../env/tryRequire.js");
var _verify = require("./verify.js");
var _check = require("./check.js");
var _error = require("./error.js");
var _time = require("./time.js");
var _format = require("./format.js");
var _string = require("./string.js");
var _function = require("./function.js");
var _Tree = require("./data/Tree.js");
const LIST_NAME_BEFORE = "beforeList";
const LIST_NAME_IT = "itList";
const LIST_NAME_AFTER = "afterList";
const createScope = (upperScope = null, level = 0, title = "") => ({
  upperScope,
  level,
  title,
  [LIST_NAME_BEFORE]: [],
  [LIST_NAME_IT]: [],
  [LIST_NAME_AFTER]: []
});
const walkScopeAsync = (0, _Tree.createTreeDepthFirstSearchAsync)((scope => scope.func ? [] : [ ...scope.beforeList, ...scope.itList, ...scope.afterList ]));
const getTitleStack = (title, upperScope) => {
  const titleStack = [ title ];
  for (;upperScope; ) {
    titleStack.unshift(upperScope.title);
    upperScope = upperScope.upperScope;
  }
  return titleStack;
};
exports.createTest = ({colorFallback = text => text, colorError = colorFallback, colorMain = colorFallback, colorTitle = colorFallback, colorText = colorFallback, colorTime = colorFallback} = {}) => {
  let ROOT_SCOPE;
  let CURRENT_SCOPE;
  let RESULT;
  let CONFIG;
  const describe = (title, setupFunc) => {
    (0, _verify.string)(title, "invalid describe title");
    (0, _verify.basicFunction)(setupFunc, "invalid describe setupFunc");
    if (void 0 === CURRENT_SCOPE) throw new Error(`should add "describe()" between "TEST_SETUP()" & "TEST_RUN()": ${title}`);
    const innerScope = createScope(CURRENT_SCOPE, CURRENT_SCOPE.level + 1, title);
    CURRENT_SCOPE.itList.push(innerScope);
    CURRENT_SCOPE = innerScope;
    const {result, error} = (0, _error.catchSync)(setupFunc);
    if (result instanceof Promise) throw new Error(`unexpect Promise from describe: ${title}`);
    if (error) {
      CONFIG.log(colorError(`error from describe: ${title}`));
      throw error;
    }
    CURRENT_SCOPE = CURRENT_SCOPE.upperScope;
  };
  const getPushScopeFunc = (type, typeListName) => (title, func) => {
    void 0 === func && (0, _check.isBasicFunction)(title) && ([title, func] = [ func, title ]);
    (0, _check.isString)(title) || (title = `${CURRENT_SCOPE.title} - ${type} #${CURRENT_SCOPE[typeListName].length}`);
    (0, _verify.string)(title, `invalid ${type} title`);
    (0, _verify.basicFunction)(func, `invalid ${type} func`);
    if (void 0 === CURRENT_SCOPE) throw new Error(`should "${type}()" between "TEST_SETUP()" & "TEST_RUN()": ${title}`);
    CURRENT_SCOPE[typeListName].push({
      upperScope: CURRENT_SCOPE,
      level: CURRENT_SCOPE.level + 1,
      title,
      func
    });
  };
  const before = getPushScopeFunc("before", "beforeList");
  const it = getPushScopeFunc("it", "itList");
  const after = getPushScopeFunc("after", "afterList");
  const stringify = ((0, _tryRequire.tryRequire)("node:util") || {}).inspect || String;
  const info = (...args) => CONFIG && CONFIG.log((0, _string.indentLine)(args.map((v => (0, _check.isString)(v) ? v : stringify(v))).join(" "), "      > "));
  return {
    TEST_SETUP: ({log = console.log, logLevel = (level, ...args) => log((0, _string.indentLine)(args.join(" "), "  ".repeat(level))), timeout = 42e3, testGlobal = globalThis, isSkipGlobalAssign = !1} = {}) => {
      ROOT_SCOPE = CURRENT_SCOPE = createScope(null, 0, "root");
      RESULT = {
        passList: [],
        failList: []
      };
      CONFIG = {
        log,
        logLevel,
        timeout
      };
      !isSkipGlobalAssign && Object.assign(testGlobal, {
        describe,
        it,
        before,
        after,
        info
      });
      CONFIG.log(colorMain("[TEST] setup"));
    },
    TEST_RUN: async () => {
      CONFIG.log(colorMain("[TEST] run"));
      const scope = ROOT_SCOPE;
      ROOT_SCOPE = CURRENT_SCOPE = void 0;
      const runStart = (0, _time.clock)();
      await walkScopeAsync(scope, (async ({upperScope, title, level, func}) => {
        if (func) {
          const funcStart = (0, _time.clock)();
          const {error} = await (0, _error.catchAsync)(_function.withTimeoutAsync, func, CONFIG.timeout);
          const funcDuration = (0, _time.clock)() - funcStart;
          const timeLog = funcDuration > 64 ? colorTime(`(${(0, _format.time)(funcDuration)})`) : "";
          if (error) {
            RESULT.failList.push({
              titleStack: getTitleStack(title, upperScope),
              error
            });
            CONFIG.logLevel(level, colorError(`[FAIL] ${title}`), timeLog);
            CONFIG.logLevel(level, colorError((0, _string.indentLine)(error.stack || error, "    ")));
          } else {
            RESULT.passList.push({
              title
            });
            CONFIG.logLevel(level, colorText(title), timeLog);
          }
        } else CONFIG.logLevel(level, colorTitle(title));
      }));
      CONFIG.log(colorMain([ `[TEST] done in ${(0, _format.time)((0, _time.clock)() - runStart)}`, `  pass: ${RESULT.passList.length}`, `  fail: ${RESULT.failList.length}` ].join("\n")));
      for (let index = 0, indexMax = RESULT.failList.length; index < indexMax; index++) {
        const {titleStack, error} = RESULT.failList[index];
        CONFIG.log(colorError(`[FAIL|${index}]`));
        titleStack.forEach(((title, level) => level && CONFIG.logLevel(level, colorError(title))));
        CONFIG.logLevel(titleStack.length, colorError((0, _string.indentLine)(error.stack || error, "    ")));
      }
      return RESULT;
    },
    describe,
    it,
    before,
    after,
    info
  };
};