var _TimedLookup = require("../../common/module/TimedLookup.js");
var _CacheMap = require("../../common/data/CacheMap.js");
var _format = require("../../common/format.js");
var _string = require("../../common/string.js");
var _EntityTag = require("./EntityTag.js");
var _net = require("../net.js");
var _File = require("../fs/File.js");
var _Path = require("../fs/Path.js");
const saveAuthFile = async (pathFile, timedLookupData) => (0, _File.writeArrayBuffer)(pathFile, (0, _TimedLookup.packDataArrayBuffer)(timedLookupData));
exports.saveAuthFile = saveAuthFile;
const loadAuthFile = async pathFile => (0, _TimedLookup.parseDataArrayBuffer)(await (0, _File.readArrayBuffer)(pathFile));
exports.loadAuthFile = loadAuthFile;
exports.describeAuthFile = async pathFile => {
  const {tag, size, tokenSize, timeGap, info, dataView} = await loadAuthFile(pathFile);
  return (0, _string.indentList)("[AuthFile]", [ `tag: ${tag}`, `size: ${(0, _format.binary)(size)}B`, `tokenSize: ${tokenSize}B`, `timeGap: ${(0, _format.time)(1e3 * timeGap)}`, info && `info: ${(0, _format.prettyStringifyJSON)(info)}`, `dataViewEntityTag: ${(0, 
  _EntityTag.getEntityTagByContentHash)(Buffer.from(dataView.buffer))}` ].filter(Boolean));
};
exports.generateAuthFile = async (pathFile, {tag, size, tokenSize, timeGap, info}) => {
  const timedLookupData = await (0, _TimedLookup.generateLookupData)({
    tag,
    size,
    tokenSize,
    timeGap,
    info
  });
  await saveAuthFile(pathFile, timedLookupData);
  return timedLookupData;
};
exports.generateAuthCheckCode = async (pathFile, timestamp) => (0, _TimedLookup.generateCheckCode)(await loadAuthFile(pathFile), timestamp);
exports.verifyAuthCheckCode = async (pathFile, checkCode, timestamp) => (0, _TimedLookup.verifyCheckCode)(await loadAuthFile(pathFile), checkCode, timestamp);
const DEFAULT_AUTH_KEY = exports.DEFAULT_AUTH_KEY = "auth-check-code";
const AUTH_SKIP = exports.AUTH_SKIP = "auth-skip";
const AUTH_FILE = exports.AUTH_FILE = "auth-file";
const AUTH_FILE_GROUP = exports.AUTH_FILE_GROUP = "auth-file-group";
const authFetchWrap = async (url, option = {}, timedLookupData, authKey) => {
  const response = await (0, _net.fetchLikeRequest)(url, {
    ...option,
    headers: {
      [authKey]: (0, _TimedLookup.generateCheckCode)(timedLookupData),
      ...option.headers
    }
  });
  if (!response.ok) throw new Error(`[authFetch] status: ${response.status}`);
  return response;
};
const configureAuthSkip = async ({authKey = DEFAULT_AUTH_KEY, log}) => {
  log && log("auth skip enabled");
  return {
    authMode: AUTH_SKIP,
    authKey,
    authFetch: _net.fetchLikeRequest,
    checkAuth: async () => "auth-skip",
    generateAuthCheckCode: async () => "auth-skip"
  };
};
exports.configureAuthSkip = configureAuthSkip;
const configureAuthFile = async ({authKey = DEFAULT_AUTH_KEY, log, timedLookupData, authFile}) => {
  if (!timedLookupData) {
    timedLookupData = await loadAuthFile(authFile).catch((async error => {
      log && log("missing auth file", error);
      throw error;
    }));
    log && log("loaded auth file");
  }
  return {
    authMode: AUTH_FILE,
    authKey,
    authFetch: async (url, option) => authFetchWrap(url, option, timedLookupData, authKey),
    checkAuth: async checkCode => {
      (0, _TimedLookup.verifyCheckCode)(timedLookupData, checkCode);
      return timedLookupData.tag;
    },
    generateAuthCheckCode: async () => (0, _TimedLookup.generateCheckCode)(timedLookupData)
  };
};
exports.configureAuthFile = configureAuthFile;
const configureAuthFileGroup = async ({authKey = DEFAULT_AUTH_KEY, log, authFileGroupPath, authFileGroupDefaultTag, authFileGroupKeySuffix, getFileNameForTag = authFileGroupKeySuffix ? tag => `${tag}${authFileGroupKeySuffix}` : tag => `${tag}.key`, authCacheExpireTime = 3e5, authCacheMap = (0, 
_CacheMap.createCacheMap)({
  valueSizeSumMax: 8388608,
  expireAfter: authCacheExpireTime,
  eventHub: null
})}) => {
  const getPath = (0, _Path.createPathPrefixLock)(authFileGroupPath);
  const getTimedLookupData = tag => authCacheMap.get(tag) || loadAuthFile(getPath(getFileNameForTag(tag))).then((timedLookupData => {
    authCacheMap.set(tag, timedLookupData, timedLookupData.dataView.byteLength);
    log && log(`loaded auth file for tag: ${tag}`);
    return timedLookupData;
  }), (error => {
    log && log(`no auth file for tag: ${tag}`);
    throw error;
  }));
  return {
    authMode: AUTH_FILE_GROUP,
    authKey,
    authFetch: async (url, option, tag = authFileGroupDefaultTag) => authFetchWrap(url, option, await getTimedLookupData(tag), authKey),
    checkAuth: async checkCode => {
      const parsedCheckCode = (0, _TimedLookup.parseCheckCode)(checkCode);
      const tag = parsedCheckCode[0];
      const timedLookupData = await getTimedLookupData(tag);
      (0, _TimedLookup.verifyParsedCheckCode)(timedLookupData, parsedCheckCode);
      return timedLookupData.tag;
    },
    generateAuthCheckCode: async (tag = authFileGroupDefaultTag) => (0, _TimedLookup.generateCheckCode)(await getTimedLookupData(tag))
  };
};
exports.configureAuthFileGroup = configureAuthFileGroup;
exports.configureAuth = async ({authKey = DEFAULT_AUTH_KEY, log, authSkip = !1, authFile, authFileGroupPath, authFileGroupDefaultTag, authFileGroupKeySuffix}) => {
  if (authSkip) return configureAuthSkip({
    authKey,
    log
  });
  if (authFile) return configureAuthFile({
    authFile,
    authKey,
    log
  });
  if (authFileGroupPath) return configureAuthFileGroup({
    authFileGroupPath,
    authFileGroupDefaultTag,
    authFileGroupKeySuffix,
    authKey,
    log
  });
  throw new Error("no auth mode option provided");
};