var _nodePath = require("node:path");
var _nodeFs = require("node:fs");
var _error = require("../../common/error.js");
var _Stream = require("../data/Stream.js");
var _Path = require("../fs/Path.js");
var _Directory = require("../fs/Directory.js");
const KEY_CONTENT_COMPACT = "cc";
const TYPE_FILE = exports.TYPE_FILE = "f";
const TYPE_DIRECTORY = exports.TYPE_DIRECTORY = "d";
const TYPE_SYMLINK = exports.TYPE_SYMLINK = "s";
const isExecutableToFileMode = isExecutable => isExecutable ? 493 : 420;
const toRoute = (packRoot, path) => (0, _Path.toPosixPath)((0, _nodePath.relative)(packRoot, path));
const autoOpenFsPack = async (asyncFunc, fsPack, openFlag, extra) => {
  const {packPath, packFh} = fsPack;
  if (packFh) return asyncFunc(packFh, fsPack, extra);
  const packFhAuto = await _nodeFs.promises.open(packPath, openFlag);
  const {result, error} = await (0, _error.catchAsync)(asyncFunc, packFhAuto, fsPack, extra);
  await packFhAuto.close();
  if (error) throw error;
  return result;
};
const readBuffer = async (packFh, offset, size, buffer = Buffer.allocUnsafe(size)) => {
  await packFh.read(buffer, 0, size, offset);
  return buffer;
};
const readHeaderNumberList = async (packFh, offset, count = 1) => {
  const buffer = await readBuffer(packFh, offset, 4 * count);
  const numberList = [];
  for (let index = 0; index < count; index++) numberList.push(buffer.readUInt32BE(4 * index));
  return numberList;
};
const writeBuffer = async (packFh, offset, buffer) => packFh.write(buffer, 0, buffer.length, offset);
const writeHeaderNumberList = async (packFh, offset, numberList) => {
  const buffer = Buffer.allocUnsafe(4 * numberList.length);
  for (let index = 0; index < numberList.length; index++) buffer.writeUInt32BE(numberList[index], 4 * index);
  return writeBuffer(packFh, offset, buffer);
};
const readFsPackHeader = async (packFh, fsPack) => {
  const {offset} = fsPack;
  const [headerVersion, headerOffset] = await readHeaderNumberList(packFh, offset, 2);
  if (headerVersion >> 8 != 4608848) throw new Error(`not fsPack: ${headerVersion.toString(36)}, expect: ${1179865089..toString(36)}`);
  if (1 != (255 & headerVersion)) throw new Error("mismatch version: " + (255 & headerVersion));
  if (!(headerOffset >= 8 && headerOffset <= 4294967295)) throw new Error(`invalid headerOffset: ${headerOffset}`);
  const [headerLength] = await readHeaderNumberList(packFh, offset + headerOffset, 1);
  const headerJSON = (buffer => {
    const {[KEY_CONTENT_COMPACT]: contentCompact = ""} = JSON.parse(String(buffer));
    return {
      contentList: contentCompact ? contentCompact.split("\r").map((contentCompact => {
        const contentCompactList = contentCompact.split("\n");
        const typeCompactList = contentCompactList[0].split(" ");
        switch (typeCompactList[0]) {
         case TYPE_FILE:
          {
            const [, routeJSON] = contentCompactList;
            const [type, sizeBase36, isExecutableMark] = typeCompactList;
            return {
              type,
              route: JSON.parse(routeJSON),
              size: parseInt(sizeBase36, 36),
              isExecutable: "E" === isExecutableMark
            };
          }
         case TYPE_DIRECTORY:
          {
            const [, routeJSON] = contentCompactList;
            const [type] = typeCompactList;
            return {
              type,
              route: JSON.parse(routeJSON)
            };
          }
         case TYPE_SYMLINK:
          {
            const [, routeJSON, targetJSON] = contentCompactList;
            const [type] = typeCompactList;
            return {
              type,
              route: JSON.parse(routeJSON),
              target: JSON.parse(targetJSON)
            };
          }
         default:
          throw new Error(`unsupported content type: ${typeCompactList[0]} ${JSON.stringify({
            contentCompactList,
            typeCompactList
          })}`);
        }
      })) : []
    };
  })(await readBuffer(packFh, offset + headerOffset + 4, headerLength));
  fsPack.headerOffset = headerOffset;
  fsPack.headerJSON = headerJSON;
};
const writeFsPackHeader = async (packFh, fsPack) => {
  const {offset, headerOffset, headerJSON} = fsPack;
  await writeHeaderNumberList(packFh, offset, [ 1179865089, headerOffset ]);
  const headerJSONBuffer = (headerJSON => {
    const {contentList} = headerJSON;
    const contentCompact = contentList.map((content => {
      switch (content.type) {
       case TYPE_FILE:
        {
          const {type, route, size, isExecutable} = content;
          return [ [ type, size.toString(36), isExecutable ? "E" : "" ].join(" ").trim(), JSON.stringify(route) ].join("\n");
        }
       case TYPE_DIRECTORY:
        {
          const {type, route} = content;
          return [ type, JSON.stringify(route) ].join("\n");
        }
       case TYPE_SYMLINK:
        {
          const {type, route, target} = content;
          return [ type, JSON.stringify(route), JSON.stringify(target) ].join("\n");
        }
       default:
        throw new Error(`unsupported content type: ${content.type}`);
      }
    })).join("\r");
    return Buffer.from(JSON.stringify({
      [KEY_CONTENT_COMPACT]: contentCompact
    }));
  })(headerJSON);
  await writeHeaderNumberList(packFh, offset + headerOffset, [ headerJSONBuffer.length ]);
  await writeBuffer(packFh, offset + headerOffset + 4, headerJSONBuffer);
};
const writeFsPackAppendFile = async (packFh, fsPack, {path, route = toRoute(fsPack.packRoot, path), size, isExecutable, stat, buffer, readStream}) => {
  if (!size || !isExecutable) {
    stat || (stat = await _nodeFs.promises.stat(path));
    size || (size = stat.size);
    isExecutable || (isExecutable = (mode = stat.mode, Boolean(73 & mode)));
  }
  var mode;
  readStream || (readStream = buffer ? (0, _Stream.bufferToReadableStream)(buffer) : (0, _nodeFs.createReadStream)(path));
  const writeStream = (0, _nodeFs.createWriteStream)(fsPack.packPath, {
    fd: packFh.fd,
    start: fsPack.offset + fsPack.headerOffset,
    autoClose: !1
  });
  await (0, _Stream.quickRunletFromStream)(readStream, writeStream);
  fsPack.headerOffset += size;
  fsPack.headerJSON.contentList.push({
    type: TYPE_FILE,
    route,
    size,
    isExecutable
  });
};
const editFsPackAppendDirectory = async (fsPack, {path, route = toRoute(fsPack.packRoot, path)}) => {
  fsPack.headerJSON.contentList.push({
    type: TYPE_DIRECTORY,
    route
  });
};
const editFsPackAppendSymlink = async (fsPack, {path, route = toRoute(fsPack.packRoot, path), target}) => {
  target || (target = await _nodeFs.promises.readlink(path));
  fsPack.headerJSON.contentList.push({
    type: TYPE_SYMLINK,
    route,
    target
  });
};
const unpackFsPackFile = async (packFh, fsPack, {route, size, isExecutable, buffer, writeStream}) => {
  const {packPath, unpackPath, fastCache} = fsPack;
  const {offsetSum} = fastCache.contentMap[route];
  const path = (0, _nodePath.resolve)(unpackPath, route);
  await (0, _Directory.createDirectory)((0, _nodePath.dirname)(path));
  if (buffer) await readBuffer(packFh, offsetSum, size, buffer); else {
    const option = {
      mode: isExecutableToFileMode(isExecutable)
    };
    writeStream || size >= 65536 ? await (0, _Stream.quickRunletFromStream)((0, _nodeFs.createReadStream)(packPath, {
      fd: packFh.fd,
      start: offsetSum,
      end: offsetSum + size - 1,
      autoClose: !1
    }), writeStream || (0, _nodeFs.createWriteStream)(path, option)) : size > 0 ? await _nodeFs.promises.writeFile(path, await readBuffer(packFh, offsetSum, size), option) : await _nodeFs.promises.writeFile(path, "", option);
  }
};
const unpackFsPackDirectory = async (packFh, fsPack, {route}) => {
  const {unpackPath} = fsPack;
  const path = (0, _nodePath.resolve)(unpackPath, route);
  await (0, _Directory.createDirectory)(path);
};
const unpackFsPackSymlink = async (packFh, fsPack, {route, target}) => {
  const {unpackPath} = fsPack;
  const path = (0, _nodePath.resolve)(unpackPath, route);
  await (0, _Directory.createDirectory)((0, _nodePath.dirname)(path));
  await _nodeFs.promises.symlink(target, path);
};
const verifyEdit = fsPack => {
  if (fsPack.isReadOnly) throw new Error("no edit readOnly fsPack");
  (fsPack => {
    fsPack.fastCache = null;
  })(fsPack);
};
const verifyUnpack = fsPack => {
  if (!fsPack.unpackPath) throw new Error("expect unpackPath");
  (fsPack => {
    if (fsPack.fastCache) return;
    const contentMap = {};
    let offsetSum = fsPack.offset + 8;
    fsPack.headerJSON.contentList.forEach((content => {
      contentMap[content.route] = {
        content,
        offsetSum
      };
      offsetSum += content.size || 0;
    }));
    fsPack.fastCache = {
      contentMap,
      offsetSum
    };
  })(fsPack);
};
exports.initFsPack = async ({packPath, packRoot = (0, _nodePath.dirname)(packPath), packFh = null, offset = 0}) => {
  const fsPack = {
    packPath,
    packRoot,
    packFh,
    offset,
    isReadOnly: !1,
    headerOffset: 8,
    headerJSON: {
      contentList: []
    },
    unpackPath: "",
    fastCache: null
  };
  await autoOpenFsPack(writeFsPackHeader, fsPack, "w+");
  return fsPack;
};
exports.saveFsPack = async fsPack => {
  verifyEdit(fsPack);
  await autoOpenFsPack(writeFsPackHeader, fsPack, "r+");
};
exports.loadFsPack = async ({packPath, packRoot = (0, _nodePath.dirname)(packPath), packFh = null, offset = 0, isReadOnly = !1, unpackPath = ""}) => {
  const fsPack = {
    packPath,
    packRoot,
    packFh,
    offset,
    isReadOnly,
    headerOffset: 0,
    headerJSON: null,
    unpackPath,
    fastCache: null
  };
  await autoOpenFsPack(readFsPackHeader, fsPack, isReadOnly ? "r" : "r+");
  return fsPack;
};
exports.setFsPackPackRoot = (fsPack, packRoot) => {
  fsPack.packRoot = packRoot;
};
const setFsPackUnpackPath = (fsPack, unpackPath) => {
  fsPack.unpackPath = unpackPath;
};
exports.setFsPackUnpackPath = setFsPackUnpackPath;
const appendFile = async (fsPack, content) => {
  verifyEdit(fsPack);
  await autoOpenFsPack(writeFsPackAppendFile, fsPack, "r+", content);
};
exports.appendFile = appendFile;
const appendDirectory = async (fsPack, content) => {
  verifyEdit(fsPack);
  await editFsPackAppendDirectory(fsPack, content);
};
exports.appendDirectory = appendDirectory;
const appendSymlink = async (fsPack, content) => {
  verifyEdit(fsPack);
  await editFsPackAppendSymlink(fsPack, content);
};
exports.appendSymlink = appendSymlink;
exports.append = async (fsPack, content) => {
  switch (content.type) {
   case TYPE_FILE:
    return appendFile(fsPack, content);
   case TYPE_DIRECTORY:
    return appendDirectory(fsPack, content);
   case TYPE_SYMLINK:
    return appendSymlink(fsPack, content);
   default:
    throw new Error(`invalid content type: ${content.type}`);
  }
};
const appendContentList = async (fsPack, contentList) => {
  verifyEdit(fsPack);
  await autoOpenFsPack((async (packFh, fsPack) => {
    for (const content of contentList) switch (content.type) {
     case TYPE_FILE:
      await writeFsPackAppendFile(packFh, fsPack, content);
      break;
     case TYPE_DIRECTORY:
      await editFsPackAppendDirectory(fsPack, content);
      break;
     case TYPE_SYMLINK:
      await editFsPackAppendSymlink(fsPack, content);
      break;
     default:
      throw new Error(`invalid content type: ${content.type}`);
    }
  }), fsPack, "r+");
};
exports.appendContentList = appendContentList;
exports.appendFromPath = async (fsPack, inputPath) => {
  verifyEdit(fsPack);
  await appendContentList(fsPack, await (async inputPath => {
    const contentList = [];
    const directoryPathList = [];
    const fileDirectorySet = new Set;
    await (0, _Directory.walkDirInfoTree)(await (0, _Directory.getDirInfoTree)(inputPath), (({type, path}) => {
      switch (type) {
       case _Path.PATH_TYPE.File:
        contentList.push({
          type: TYPE_FILE,
          path
        });
        fileDirectorySet.add((0, _nodePath.dirname)(path));
        break;
       case _Path.PATH_TYPE.Directory:
        directoryPathList.push(path);
        break;
       case _Path.PATH_TYPE.Symlink:
        contentList.push({
          type: TYPE_SYMLINK,
          path
        });
        fileDirectorySet.add((0, _nodePath.dirname)(path));
        break;
       default:
        throw new Error(`unsupported content type: ${type}, from: ${path}`);
      }
    }));
    const skipDirectorySet = new Set;
    for (let fileDirectory of fileDirectorySet) {
      let prevFileDirectory = "";
      for (;fileDirectory !== prevFileDirectory; ) {
        skipDirectorySet.add(fileDirectory);
        prevFileDirectory = fileDirectory;
        fileDirectory = (0, _nodePath.dirname)(fileDirectory);
      }
    }
    for (const path of directoryPathList) skipDirectorySet.has(path) || contentList.push({
      type: TYPE_DIRECTORY,
      path
    });
    return contentList;
  })(inputPath));
};
const unpackFile = async (fsPack, content) => {
  verifyUnpack(fsPack);
  await autoOpenFsPack(unpackFsPackFile, fsPack, "r+", content);
};
exports.unpackFile = unpackFile;
const unpackDirectory = async (fsPack, content) => {
  verifyUnpack(fsPack);
  await autoOpenFsPack(unpackFsPackDirectory, fsPack, "r+", content);
};
exports.unpackDirectory = unpackDirectory;
const unpackSymlink = async (fsPack, content) => {
  verifyUnpack(fsPack);
  await autoOpenFsPack(unpackFsPackSymlink, fsPack, "r+", content);
};
exports.unpackSymlink = unpackSymlink;
exports.unpack = async (fsPack, content) => {
  switch (content.type) {
   case TYPE_FILE:
    return unpackFile(fsPack, content);
   case TYPE_DIRECTORY:
    return unpackDirectory(fsPack, content);
   case TYPE_SYMLINK:
    return unpackSymlink(fsPack, content);
   default:
    throw new Error(`invalid content type: ${content.type}`);
  }
};
const unpackContentList = async (fsPack, contentList) => {
  verifyUnpack(fsPack);
  await autoOpenFsPack((async (packFh, fsPack) => {
    for (const content of contentList) switch (content.type) {
     case TYPE_FILE:
      await unpackFsPackFile(packFh, fsPack, content);
      break;
     case TYPE_DIRECTORY:
      await unpackFsPackDirectory(0, fsPack, content);
      break;
     case TYPE_SYMLINK:
      await unpackFsPackSymlink(0, fsPack, content);
      break;
     default:
      throw new Error(`invalid content type: ${content.type}`);
    }
  }), fsPack, "r+");
};
exports.unpackContentList = unpackContentList;
exports.unpackToPath = async (fsPack, unpackPath) => {
  unpackPath && setFsPackUnpackPath(fsPack, unpackPath);
  await unpackContentList(fsPack, fsPack.headerJSON.contentList);
};