var _nodeOs = require("node:os");
var _nodeV = require("node:v8");
var _format = require("../../common/format.js");
var _string = require("../../common/string.js");
var _File = require("../fs/File.js");
const getSystemPlatform = () => ({
  platform: (0, _nodeOs.platform)(),
  arch: (0, _nodeOs.arch)(),
  release: (0, _nodeOs.release)()
});
exports.getSystemPlatform = getSystemPlatform;
const describeSystemPlatform = ({platform, arch, release} = getSystemPlatform()) => `${platform} ${arch} [${release}]`;
exports.describeSystemPlatform = describeSystemPlatform;
const getSystemProcessor = () => (0, _nodeOs.cpus)() || [ {
  model: "N/A",
  speed: "N/A",
  times: {
    user: 0,
    nice: 0,
    sys: 0,
    idle: 0,
    irq: 0
  }
} ];
exports.getSystemProcessor = getSystemProcessor;
const describeSystemProcessor = (processorList = getSystemProcessor()) => processorList.map((({model, speed, times}) => `[${model}] speed:${speed}MHz ${Object.entries(times).map((([k, v]) => `${k}:${(0, _format.time)(v)}`)).join(" ")}`)).join("\n");
exports.describeSystemProcessor = describeSystemProcessor;
const getSystemMemory = () => {
  try {
    if ("linux" === (0, _nodeOs.platform)()) {
      const stringMemInfo = (0, _File.readTextSync)("/proc/meminfo");
      const toByte = regexp => 1024 * parseInt(regexp.exec(stringMemInfo)[1]);
      return {
        total: toByte(/MemTotal:\s+(\d+)/),
        free: toByte(/MemAvailable:\s+(\d+)/),
        swapTotal: toByte(/SwapTotal:\s+(\d+)/),
        swapFree: toByte(/SwapFree:\s+(\d+)/)
      };
    }
  } catch (error) {}
  return {
    total: (0, _nodeOs.totalmem)(),
    free: (0, _nodeOs.freemem)(),
    swapTotal: 0,
    swapFree: 0
  };
};
exports.getSystemMemory = getSystemMemory;
const describeSystemMemory = ({total, free, swapTotal, swapFree} = getSystemMemory()) => [ `Used: ${(0, _format.percent)((total - free) / total)}${swapTotal ? `+${(0, _format.percent)((swapTotal - swapFree) / swapTotal)}` : ""}`, `Total: ${(0, 
_format.binary)(total)}B${swapTotal ? `+${(0, _format.binary)(swapTotal)}B` : ""}`, `Free: ${(0, _format.binary)(free)}B${swapTotal ? `+${(0, _format.binary)(swapFree)}B` : ""}` ].join("\n");
exports.describeSystemMemory = describeSystemMemory;
const getSystemNetwork = () => ({
  hostname: (0, _nodeOs.hostname)(),
  networkInterface: (0, _nodeOs.networkInterfaces)()
});
exports.getSystemNetwork = getSystemNetwork;
const describeSystemNetwork = ({hostname, networkInterface} = getSystemNetwork()) => [ `[hostname] ${hostname}`, ...Object.entries(networkInterface).map((([name, addressList]) => (0, _string.indentList)(`[interface] ${name}`, addressList.map((({address, netmask, mac, internal, family, cidr}) => `[${family}${internal ? "|INTERNAL" : ""}] ${cidr || address} (${mac})`))))) ].join("\n");
exports.describeSystemNetwork = describeSystemNetwork;
const getSystemActivity = () => ({
  uptime: 1e3 * (0, _nodeOs.uptime)(),
  loadAverageList: (0, _nodeOs.loadavg)()
});
exports.getSystemActivity = getSystemActivity;
const describeSystemActivity = ({uptime, loadAverageList} = getSystemActivity()) => [ `[uptime] ${(0, _format.time)(uptime)}`, "win32" !== (0, _nodeOs.platform)() && `[load average] ${loadAverageList.map(_format.percent).join(", ")} (1min, 5min, 15min)` ].filter(Boolean).join("\n");
exports.describeSystemActivity = describeSystemActivity;
const getSystemStatus = () => ({
  platform: getSystemPlatform(),
  processor: getSystemProcessor(),
  memory: getSystemMemory(),
  network: getSystemNetwork(),
  activity: getSystemActivity()
});
exports.getSystemStatus = getSystemStatus;
exports.describeSystemStatus = ({platform, processor, memory, network, activity} = getSystemStatus()) => Object.entries({
  platform: describeSystemPlatform(platform),
  processor: describeSystemProcessor(processor),
  memory: describeSystemMemory(memory),
  network: describeSystemNetwork(network),
  activity: describeSystemActivity(activity)
}).map((([k, v]) => `[${k}]\n${(0, _string.indentLine)(v)}`)).join("\n");
exports.getSystemInfo = () => {
  const toGiB = byte => Math.round(byte / 2 ** 30);
  const toPercent = (total, free) => total ? `${(100 * (1 - free / total)).toFixed(0)}%` : "-";
  const {total, free, swapTotal, swapFree} = getSystemMemory();
  const nCPU = (0, _nodeOs.cpus)().length;
  const nMEM = toGiB(total);
  const nSWAP = toGiB(swapTotal);
  const loadAvgList = (0, _nodeOs.loadavg)();
  const systemStatus = [ swapTotal ? `[${nCPU}CPU|${nMEM}+${nSWAP}G]` : `[${nCPU}CPU|${nMEM}G]`, swapTotal ? `mem/swap=${toPercent(total, free)}/${toPercent(swapTotal, swapFree)}` : `mem=${toPercent(total, free)}`, `loadavg=${loadAvgList.map((v => v.toFixed(1))).join("|")}`, `uptime=${(0, 
  _format.time)(1e3 * (0, _nodeOs.uptime)())}` ].join(" ");
  const [load1Min, load5Min, load15Min] = loadAvgList;
  return {
    nCPU,
    nMEM,
    nSWAP,
    systemStatus,
    loadStatus: load15Min > 1.1 * nCPU && load5Min > .7 * nCPU && load1Min > .7 * nCPU ? "15MIN" : load5Min > 1.3 * nCPU && load1Min > .7 * nCPU ? "5MIN" : load5Min > .7 * nCPU && load1Min > 1.7 * nCPU ? "1MIN" : ""
  };
};
const V8_HEAP_RESERVED_SIZE = exports.V8_HEAP_RESERVED_SIZE = 1024 * ((0, _nodeOs.totalmem)() < 2147483648 ? 24 : 48) * 1024;
exports.getV8HeapStatus = () => {
  const {used_heap_size: v8HeapUsed, heap_size_limit: v8HeapMax} = (0, _nodeV.getHeapStatistics)();
  const total = v8HeapMax - V8_HEAP_RESERVED_SIZE;
  return {
    total,
    free: Math.max(total - v8HeapUsed, 0)
  };
};