var _nodePath = require("node:path");
var _format = require("../../common/format.js");
var _check = require("../../common/check.js");
var _verify = require("../../common/verify.js");
var _time = require("../../common/time.js");
var _LoopIndex = require("../../common/data/LoopIndex.js");
var _File = require("../fs/File.js");
var _Directory = require("../fs/Directory.js");
var _docker = require("../module/Software/docker.js");
var _Process = require("./Process.js");
var _run = require("../run.js");
const SAMPLE_COMMAND = {
  argList: "node -e process.version",
  logFile: "path/to/redirect/stdout&stderr",
  cwd: "path/to/run/command",
  env: {},
  wait: 256,
  func: async (unitConfig, ...args) => {}
};
const SAMPLE_UNIT_CONFIG = {
  name: "sample unit",
  clue: {
    pid: !0,
    pidFile: "path/to/file.pid",
    pidFileKeepStale: !1,
    commandPattern: "string-to-match",
    containerImagePattern: "string-to-match",
    containerNamesPattern: "string-to-match"
  },
  run: {
    isMajorUnit: !1,
    start: SAMPLE_COMMAND,
    stop: SAMPLE_COMMAND,
    reset: SAMPLE_COMMAND,
    resetSub: SAMPLE_COMMAND
  },
  limit: {
    memoryMaxTotalMiB: 0,
    memoryMaxSubProcessMiB: 0
  }
};
const formatPattern = pattern => {
  if ((0, _check.isString)(pattern)) return string => string.includes(pattern);
  if ((0, _check.isBasicFunction)(pattern.test)) return string => pattern.test(string);
  throw new Error(`invalid pattern: ${(0, _format.describe)(pattern)}`);
};
const formatCommand = (command, formatPath, isDetached = !1) => {
  if (command.argList) {
    let {argList, logFile, cwd, env, wait = 256} = command;
    (0, _check.isString)(argList) && (argList = argList.split(" "));
    (0, _verify.basicArray)(argList, `invalid command.argList: ${(0, _format.describe)(argList)}`);
    logFile = logFile ? formatPath(logFile) : void 0;
    cwd = cwd ? formatPath(cwd) : void 0;
    void 0 !== env && (0, _verify.basicObject)(env, `invalid command.env: ${(0, _format.describe)(env)}`);
    (0, _verify.integer)(wait, `invalid command.wait: ${(0, _format.describe)(wait)}`);
    return async (unitConfig, ...args) => {
      const option = {
        stdoutFile: isDetached ? logFile : void 0,
        env: env && {
          ...process.env,
          ...env
        },
        cwd
      };
      let pid;
      if (isDetached) pid = (0, _run.runDetached)(argList, option).subProcess.pid; else {
        const {subProcess, promise} = (0, _run.run)(argList, option);
        await promise.catch((error => {}));
        pid = subProcess.pid;
      }
      wait && await (0, _time.setTimeoutAsync)(wait);
      return pid;
    };
  }
  if (command.func) {
    (0, _verify.basicFunction)(command.func, `invalid command.func: ${(0, _format.describe)(command.func)}`);
    return command.func;
  }
  throw new Error(`invalid command: ${(0, _format.describe)(command)}`);
};
const defaultCommandStop = async (unitConfig, ...args) => {
  await (0, _Process.killProcessInfoAsync)({
    pid: args[0]
  });
};
exports.defaultCommandStop = defaultCommandStop;
const formatUnitConfig = ({name, clue = {
  pid: !0
}, run, limit = {}}, formatPath) => {
  (0, _verify.string)(name, "invalid unitConfig.name");
  (0, _verify.basicObject)(clue, `invalid unitConfig.clue for ${name}`);
  if (clue.pid) clue = {
    pid: Boolean(clue.pid)
  }; else if (clue.pidFile) clue = {
    pidFile: formatPath(clue.pidFile)
  }; else if (clue.commandPattern) clue = {
    commandPattern: formatPattern(clue.commandPattern)
  }; else if (clue.containerImagePattern) clue = {
    containerImagePattern: formatPattern(clue.containerImagePattern)
  }; else {
    if (!clue.containerNamesPattern) throw new Error(`no supported unitConfig.clue for ${name}`);
    clue = {
      containerNamesPattern: formatPattern(clue.containerNamesPattern)
    };
  }
  const isCheckDocker = clue.containerImagePattern || clue.containerNamesPattern;
  (0, _verify.basicObject)(run, `invalid unitConfig.run for ${name}`);
  const __run = {
    isMajorUnit: Boolean(run.isMajorUnit),
    start: formatCommand(run.start, formatPath, "run-detached"),
    stop: run.stop ? formatCommand(run.stop, formatPath) : defaultCommandStop,
    reset: run.reset || run.stop ? formatCommand(run.reset || run.stop, formatPath) : defaultCommandStop,
    resetSub: run.resetSub || run.reset ? formatCommand(run.resetSub || run.reset, formatPath) : defaultCommandStop
  };
  (0, _verify.basicObject)(limit, `invalid unitConfig.limit for ${name}`);
  return {
    name,
    clue,
    __run,
    limit: limit = {
      memoryMaxTotalMiB: limit.memoryMaxTotalMiB || 0,
      memoryMaxSubProcessMiB: limit.memoryMaxSubProcessMiB || 0
    },
    isCheckDocker
  };
};
exports.formatUnitConfig = formatUnitConfig;
exports.formatLoopConfig = ({configRoot = process.cwd(), stateFilePath = "", loopTime = 6e3, loopCheckInterval = 5, loopStartReRunInterval = 2, unitConfigList = [] || [ SAMPLE_UNIT_CONFIG ], isCheckDocker = !1}) => {
  const formatPath = (...path) => (0, _nodePath.resolve)(configRoot, ...path);
  stateFilePath = formatPath(stateFilePath);
  unitConfigList = unitConfigList.map((unitConfig => formatUnitConfig(unitConfig, formatPath)));
  isCheckDocker = Boolean(isCheckDocker || unitConfigList.some((unitConfig => unitConfig.isCheckDocker)));
  const nameSet = new Set;
  for (const {name} of unitConfigList) {
    if (nameSet.has(name)) throw new Error(`duplicate unitConfig.name: ${name}`);
    nameSet.add(name);
  }
  return {
    configRoot,
    stateFilePath,
    loopTime,
    loopCheckInterval,
    loopStartReRunInterval,
    unitConfigList,
    isCheckDocker
  };
};
const chopListLength = list => {
  list.length > 16 && (list.length = 16);
};
const addLoopHistory = (loopHistoryList, ...noteList) => {
  loopHistoryList.unshift({
    time: Date.now(),
    note: noteList.filter(Boolean).join(" ")
  });
  chopListLength(loopHistoryList);
};
const initUnitState = () => ({
  historyList: [ {
    time: Date.now(),
    loopCount: 0,
    state: "missing",
    reason: "init"
  } ],
  latestFoundTime: 0,
  latestMissingTime: Date.now(),
  clueProcessInfo: null
});
const initLoopState = (loopConfig, loopExtraData = {}) => {
  const loopHistoryList = [];
  const unitStateMap = {};
  for (const unitConfig of loopConfig.unitConfigList) unitStateMap[unitConfig.name] = initUnitState();
  addLoopHistory(loopHistoryList, "#0 init");
  return {
    loopIndex: 0,
    loopHistoryList,
    unitStateMap,
    loopExtraData
  };
};
exports.initLoopState = initLoopState;
exports.loadLoopState = async loopConfig => {
  try {
    const {loopIndex, loopHistoryList, unitStateMap: loadUnitStateMap, loopExtraData = {}} = await (0, _File.readJSON)(loopConfig.stateFilePath);
    const unitStateMap = {};
    for (const unitConfig of loopConfig.unitConfigList) {
      const prevUnitState = loadUnitStateMap[unitConfig.name];
      unitStateMap[unitConfig.name] = prevUnitState || initUnitState();
    }
    const danglingUnitTagList = [];
    for (const [prevName, prevUnitState] of Object.entries(loadUnitStateMap)) !unitStateMap[prevName] && latestUnitStateHistory(prevUnitState).state.startsWith("found") && danglingUnitTagList.push(`${prevName}@${prevUnitState.clueProcessInfo.pid}`);
    addLoopHistory(loopHistoryList, `#${loopIndex} load`, danglingUnitTagList.length && `dangling: ${danglingUnitTagList.join(",")}`);
    return {
      loopIndex,
      loopHistoryList,
      unitStateMap,
      loopExtraData
    };
  } catch (error) {
    return initLoopState(loopConfig);
  }
};
exports.saveLoopState = async (loopConfig, loopState) => {
  await (0, _Directory.createDirectory)((0, _nodePath.dirname)(loopConfig.stateFilePath));
  await (0, _File.writeJSON)(loopConfig.stateFilePath, loopState);
};
exports.migrateLoopState = async (loopConfig, prevLoopConfig, prevLoopState) => {
  prevLoopState = (loopConfig => JSON.parse(JSON.stringify(loopConfig)))(prevLoopState);
  const {loopIndex, loopHistoryList, loopExtraData = {}} = prevLoopState;
  const unitStateMap = {};
  for (const unitConfig of loopConfig.unitConfigList) {
    const prevUnitState = prevLoopState.unitStateMap[unitConfig.name];
    unitStateMap[unitConfig.name] = prevUnitState || initUnitState();
  }
  const droppedUnitTagList = [];
  for (const prevUnitConfig of prevLoopConfig.unitConfigList) {
    if (unitStateMap[prevUnitConfig.name]) continue;
    const prevUnitState = prevLoopState.unitStateMap[prevUnitConfig.name];
    if (latestUnitStateHistory(prevUnitState).state.startsWith("found")) {
      droppedUnitTagList.push(`${prevUnitConfig.name}@${prevUnitState.clueProcessInfo.pid}`);
      await prevUnitConfig.__run.stop(prevUnitConfig, prevUnitState.clueProcessInfo.pid);
    }
  }
  addLoopHistory(loopHistoryList, `#${loopIndex} migrate`, droppedUnitTagList.length && `drop: ${droppedUnitTagList.join(",")}`);
  return {
    loopIndex,
    loopHistoryList,
    unitStateMap,
    loopExtraData
  };
};
exports.markLoopState = (loopConfig, loopState, ...noteList) => {
  const {loopIndex, loopHistoryList} = loopState;
  addLoopHistory(loopHistoryList, `#${loopIndex}`, ...noteList);
  return loopState;
};
const latestUnitStateHistory = unitState => unitState.historyList[0];
exports.latestUnitStateHistory = latestUnitStateHistory;
const addUnitStateHistory = (unitState, state, reason = "") => {
  unitState.historyList.unshift({
    time: Date.now(),
    loopCount: 0,
    state,
    reason
  });
  chopListLength(unitState.historyList);
  "found" === state ? unitState.latestFoundTime = Date.now() : "missing" === state && (unitState.latestMissingTime = Date.now());
};
exports.addUnitStateHistory = addUnitStateHistory;
const LOOP_INDEX = exports.LOOP_INDEX = (0, _LoopIndex.createLoopIndex)();
const toMiB = byte => Math.round(byte / 2 ** 20);
exports.loopWaitAndStep = async (loopConfig, loopState) => {
  await (0, _time.setTimeoutAsync)(loopConfig.loopTime);
  loopState.loopIndex = LOOP_INDEX.step(loopState.loopIndex, 1);
  for (const unitState of Object.values(loopState.unitStateMap)) unitState.historyList[0].loopCount++;
};
const clueFind = async (unitConfig, unitState, SPI) => {
  const {clue} = unitConfig;
  let processInfo;
  if (clue.pid) {
    processInfo = clue.pid && (0, _Process.isPidExist)(clue.pid) && SPI.processPidMap[clue.pid];
    unitState.clueProcessInfo && unitState.clueProcessInfo.command !== processInfo.command && (processInfo = void 0);
  } else if (clue.pidFile) {
    const pid = parseInt((await (0, _File.readText)(clue.pidFile)).trim());
    processInfo = pid && (0, _Process.isPidExist)(pid) && SPI.processPidMap[pid];
    processInfo || clue.pidFileKeepStale || await (0, _File.deleteFile)(clue.pidFile);
  } else if (clue.commandPattern) processInfo = SPI.processList.find((({command}) => clue.commandPattern(command))); else if (clue.containerImagePattern || clue.containerNamesPattern) {
    const dockerContainer = SPI.dockerContainerList.find(clue.containerImagePattern ? ({image}) => clue.containerImagePattern(image) : ({names}) => clue.containerNamesPattern(names));
    processInfo = dockerContainer && SPI.processList.find((({pid}) => pid === dockerContainer.pid));
  }
  return processInfo;
};
const loopClue = async (loopConfig, loopState) => {
  const {unitConfigList, isCheckDocker} = loopConfig;
  const SPI = await (async ({isCheckDocker = !1}) => {
    const processList = await (0, _Process.getProcessListAsync)();
    const processPidMap = (0, _Process.toProcessPidMap)(processList);
    const processTree = (0, _Process.toProcessTree)(processList);
    const dockerContainerList = isCheckDocker ? await (0, _docker.getContainerLsList)() : [];
    (0, _docker.matchContainerLsList)(dockerContainerList, processList);
    return {
      processList,
      processPidMap,
      processTree,
      dockerContainerList
    };
  })({
    isCheckDocker
  });
  const foundUnitList = [];
  const missingUnitList = [];
  for (const unitConfig of unitConfigList) {
    const unitState = loopState.unitStateMap[unitConfig.name];
    const processInfo = await clueFind(unitConfig, unitState, SPI);
    if (processInfo) {
      unitState.clueProcessInfo && unitState.clueProcessInfo.pid !== processInfo.pid && addUnitStateHistory(unitState, "found", "clue:pid-change");
      latestUnitStateHistory(unitState).state.startsWith("found") || addUnitStateHistory(unitState, "found", "clue:found");
      unitState.clueProcessInfo = {
        pid: processInfo.pid,
        command: processInfo.command
      };
      foundUnitList.push({
        unitConfig,
        unitState,
        processInfo
      });
    } else {
      latestUnitStateHistory(unitState).state.startsWith("missing") || addUnitStateHistory(unitState, "missing", "clue:missing");
      unitState.clueProcessInfo = null;
      missingUnitList.push({
        unitConfig,
        unitState
      });
    }
  }
  return {
    SPI,
    foundUnitList,
    missingUnitList
  };
};
exports.loopClue = loopClue;
const limitCheck = async (unitConfig, processInfo) => {
  const {limit} = unitConfig;
  if (limit.memoryMaxSubProcessMiB) {
    let isLimitHit = !1;
    for (const subInfo of processInfo.subInfoList) if (!(toMiB(subInfo.memory) < limit.memoryMaxSubProcessMiB)) {
      isLimitHit = !0;
      await unitConfig.__run.resetSub(unitConfig, subInfo.pid);
    }
    if (isLimitHit) return;
  }
  if (limit.memoryMaxTotalMiB && processInfo.memoryTotalMiB >= limit.memoryMaxTotalMiB) return "memoryMaxTotalMiB";
};
const statusCheck = async (unitConfig, unitState, processInfo = {}) => {
  const {pid, subInfoList = [], memoryTotalMiB = 0} = processInfo;
  const latestState = latestUnitStateHistory(unitState).state;
  const isFound = latestState.startsWith("found");
  return [ unitConfig.name, pid || "-", (0, _format.time)(Date.now() - (isFound ? unitState.latestFoundTime : unitState.latestMissingTime)), isFound ? memoryTotalMiB ? `${memoryTotalMiB}M${subInfoList.length ? `=${[ processInfo, ...subInfoList ].map((({memory}) => toMiB(memory))).join("+")}` : ""}` : "-" : latestState ];
};
exports.loopMain = async (loopConfig, loopState, {isNoStart = !1} = {}) => {
  const {loopCheckInterval, loopStartReRunInterval} = loopConfig;
  if (loopState.__isAllFound && loopState.loopIndex % loopCheckInterval != 0) {
    for (const unitState of Object.values(loopState.unitStateMap)) latestUnitStateHistory(unitState).state.startsWith("found") && (0, _Process.isPidExist)(unitState.clueProcessInfo.pid) || (loopState.__isAllFound = !1);
    if (loopState.__isAllFound) return {};
  }
  const {SPI, foundUnitList, missingUnitList} = await loopClue(loopConfig, loopState);
  loopState.__isAllFound = !missingUnitList.length;
  const table = [];
  const hasMissingMajorUnit = missingUnitList.some((({unitConfig: {__run: {isMajorUnit}}}) => isMajorUnit));
  for (const {unitConfig, unitState} of missingUnitList) {
    if (!isNoStart) {
      "missing->found" !== latestUnitStateHistory(unitState).state && addUnitStateHistory(unitState, "missing->found", "run.start");
      unitState.historyList[0].loopCount % loopStartReRunInterval == 0 && (hasMissingMajorUnit && !unitConfig.__run.isMajorUnit || await unitConfig.__run.start(unitConfig));
    }
    table.push(await statusCheck(unitConfig, unitState));
  }
  for (const {unitConfig, unitState, processInfo} of foundUnitList) {
    processInfo.subInfoList = (0, _Process.flattenProcessTree)(processInfo);
    processInfo.memoryTotalMiB = toMiB(processInfo.memory + processInfo.subInfoList.reduce(((o, {memory}) => o + memory), 0));
    const resetReason = await limitCheck(unitConfig, processInfo);
    if (resetReason) {
      "found->missing" !== latestUnitStateHistory(unitState).state && addUnitStateHistory(unitState, "found->missing", `limit:${resetReason}`);
      unitState.historyList[0].loopCount % loopStartReRunInterval == 0 && await unitConfig.__run.reset(unitConfig, processInfo.pid);
    }
    table.push(await statusCheck(unitConfig, unitState, processInfo));
  }
  return {
    SPI,
    foundUnitList,
    missingUnitList,
    statusList: [ ...(0, _format.padTable)({
      table,
      cellPad: " ",
      rowPad: "\0"
    }).split("\0") ]
  };
};
exports.loopStop = async (loopConfig, loopState) => {
  const {SPI, foundUnitList, missingUnitList} = await loopClue(loopConfig, loopState);
  for (const {unitConfig, unitState, processInfo} of foundUnitList) {
    "found->missing" !== latestUnitStateHistory(unitState).state && addUnitStateHistory(unitState, "found->missing", "run.stop");
    0 === unitState.historyList[0].loopCount && await unitConfig.__run.stop(unitConfig, processInfo.pid);
  }
  return {
    SPI,
    foundUnitList,
    missingUnitList
  };
};