import { BASIC_EXTENSION_MAP } from "../common/module/MIME.js";
const {requestAnimationFrame, document, navigator, caches, URL, Blob, Request, Response} = window;
const throttleByAnimationFrame = func => {
  let callArgs = null;
  const frameFunc = () => {
    const currentCallArgs = callArgs;
    callArgs = null;
    func.apply(null, currentCallArgs);
  };
  return (...args) => {
    !callArgs && requestAnimationFrame(frameFunc);
    callArgs = args;
  };
};
const applyReceiveFileListListener = (eventSource = document, onFileList) => {
  const muteEvent = event => {
    event.stopPropagation();
    event.preventDefault();
  };
  const pasteListener = event => {
    const {files} = event.dataTransfer || event.clipboardData;
    files && files.length && onFileList(files);
  };
  const dropListener = event => {
    muteEvent(event);
    pasteListener(event);
  };
  eventSource.addEventListener("dragenter", muteEvent);
  eventSource.addEventListener("dragover", muteEvent);
  eventSource.addEventListener("drop", dropListener);
  eventSource.addEventListener("paste", pasteListener);
  return () => {
    eventSource.removeEventListener("dragenter", muteEvent);
    eventSource.removeEventListener("dragover", muteEvent);
    eventSource.removeEventListener("drop", dropListener);
    eventSource.removeEventListener("paste", pasteListener);
  };
};
const getPathElementList = (fromElement, toElement) => {
  if (!fromElement.contains(toElement)) return [];
  let element = toElement;
  const elementList = [];
  for (;element !== fromElement; ) {
    elementList.unshift(element);
    element = element.parentElement;
  }
  return elementList;
};
const getElementAtViewport = (clientPosition, excludeElementList) => {
  const styleRecoverList = excludeElementList && excludeElementList.map((element => {
    const {visibility} = element.style;
    element.style.visibility = "hidden";
    return visibility;
  }));
  const elementUnder = document.elementFromPoint(clientPosition.x, clientPosition.y);
  excludeElementList && excludeElementList.forEach(((element, index) => element.style.visibility = styleRecoverList[index]));
  return elementUnder;
};
const createElement = (tagName, attributeMap = {}) => Object.assign(document.createElement(tagName), attributeMap);
const createDownload = (fileName, url) => {
  const element = createElement("a", {
    download: fileName,
    href: url
  });
  document.body.appendChild(element);
  element.click();
  document.body.removeChild(element);
};
const createDownloadWithBlob = (fileName, blob) => {
  if (navigator.msSaveOrOpenBlob) return navigator.msSaveOrOpenBlob(blob, fileName);
  const objectUrl = URL.createObjectURL(blob);
  createDownload(fileName, objectUrl);
  setTimeout((() => URL.revokeObjectURL(objectUrl)), 5e3);
};
const createDownloadWithString = (fileName, string, type = BASIC_EXTENSION_MAP.txt) => createDownloadWithBlob(fileName, new Blob([ string ], {
  type
}));
const createDownloadWithObject = (fileName, object, type = BASIC_EXTENSION_MAP.json) => createDownloadWithString(fileName, JSON.stringify(object), type);
const saveArrayBufferCache = async (bucketName, key, arrayBuffer) => {
  const cache = await caches.open(bucketName);
  await cache.put(new Request(key), new Response(arrayBuffer));
  return {
    bucketName,
    key
  };
};
const loadArrayBufferCache = async (bucketName, key) => {
  const cache = await caches.open(bucketName);
  const response = await cache.match(new Request(key));
  return response && response.arrayBuffer();
};
const deleteArrayBufferCache = async (bucketName, key) => {
  if (!key) return caches.delete(bucketName);
  const cache = await caches.open(bucketName);
  return cache && cache.delete(new Request(key));
};
export { throttleByAnimationFrame, applyReceiveFileListListener, getPathElementList, getElementAtViewport, createElement, createDownload, createDownloadWithBlob, createDownloadWithString, createDownloadWithObject, saveArrayBufferCache, loadArrayBufferCache, deleteArrayBufferCache };