import { getQuickContext2d } from "./ImageData.js";
const getPixelColor = (imageData, x, y) => new Uint32Array(imageData.data.buffer)[x + y * imageData.width];
const replacePixelColor = (imageData, replacerFunc) => {
  const data = new Uint32Array(imageData.data.buffer);
  for (let index = 0, indexMax = imageData.width * imageData.height; index < indexMax; index += 4) {
    const color = replacerFunc(data[index]);
    color && (data[index] = color);
  }
};
const drawPixel = (imageData, x, y, color) => new Uint32Array(imageData.data.buffer)[x + y * imageData.width] = color;
const drawPixelLine = (imageData, point0, point1, color) => {
  let x0 = Math.round(point0.x);
  let y0 = Math.round(point0.y);
  const x1 = Math.round(point1.x);
  const y1 = Math.round(point1.y);
  const dx = Math.abs(x1 - x0);
  const dy = Math.abs(y1 - y0);
  const sx = x0 < x1 ? 1 : -1;
  const sy = y0 < y1 ? 1 : -1;
  let err = dx - dy;
  for (;;) {
    drawPixel(imageData, x0, y0, color);
    if (x0 === x1 && y0 === y1) break;
    const e2 = 2 * err;
    if (e2 > -dy) {
      err -= dy;
      x0 += sx;
    }
    if (e2 < dx) {
      err += dx;
      y0 += sy;
    }
  }
};
const drawPixelLineList = (imageData, pointList, color, isLoop) => {
  if (pointList.length <= 1) throw new Error(`[drawPixelLineList] error pointList length: ${pointList.length}`);
  let fromPoint = isLoop ? pointList[pointList.length - 1] : pointList[0];
  let pointIndex = isLoop ? 0 : 1;
  for (;pointIndex < pointList.length; ) {
    drawPixelLine(imageData, fromPoint, pointList[pointIndex], color);
    fromPoint = pointList[pointIndex];
    pointIndex++;
  }
};
const scale = (imageData, scaleX, scaleY = scaleX) => {
  const sourceImageData = imageData;
  const sourcePixelWidth = sourceImageData.width;
  const sourcePixelArray = new Uint32Array(sourceImageData.data.buffer);
  const targetImageData = getQuickContext2d().getImageData(0, 0, sourcePixelWidth * scaleX, sourceImageData.height * scaleY);
  const targetPixelWidth = targetImageData.width;
  const targetPixelArray = new Uint32Array(targetImageData.data.buffer);
  for (let index = 0, indexMax = targetPixelWidth * targetImageData.height; index < indexMax; index++) {
    const targetX = index % targetPixelWidth;
    const targetY = Math.floor(index / targetPixelWidth);
    const sourcePixelIndex = Math.floor(targetX / scaleX) + Math.floor(targetY / scaleY) * sourcePixelWidth;
    targetPixelArray[index] = sourcePixelArray[sourcePixelIndex];
  }
  return targetImageData;
};
const crop = (imageData, cropCheckFunc) => {
  const sourceWidth = imageData.width;
  const sourceHeight = imageData.height;
  const sourcePixelArray = new Uint32Array(imageData.data.buffer);
  let minX = 1 / 0;
  let minY = 1 / 0;
  let maxX = -1 / 0;
  let maxY = -1 / 0;
  for (let index = 0, indexMax = sourceWidth * sourceHeight; index < indexMax; index++) {
    if (cropCheckFunc(sourcePixelArray[index])) continue;
    const x = index % sourceWidth;
    const y = Math.floor(index / sourceWidth);
    minX = Math.min(minX, x);
    minY = Math.min(minY, y);
    maxX = Math.max(maxX, x);
    maxY = Math.max(maxY, y);
  }
  const targetWidth = maxX - minX + 1;
  const targetHeight = maxY - minY + 1;
  if (targetWidth <= 0 || targetHeight <= 0) throw new Error("[crop] invalid result size");
  const bufferCanvasContext2D = getQuickContext2d(targetWidth, targetHeight);
  bufferCanvasContext2D.putImageData(imageData, 0 - minX, 0 - minY, minX, minY, targetWidth, targetHeight);
  return bufferCanvasContext2D.getImageData(0, 0, targetWidth, targetHeight);
};
const floodFill = (imageData, {x, y}, toColor) => {
  x = Math.round(x);
  y = Math.round(y);
  const {width, height} = imageData;
  const data = new Uint32Array(imageData.data.buffer);
  const fromColor = data[y * width + x];
  fromColor !== toColor && floodFillStackLoop([ {
    x,
    y
  } ], data, width, height, fromColor, toColor);
};
const floodFillStackLoop = (markPointList, data, width, height, fromColor, toColor) => {
  const checkPixelColor = (x, y, color) => data[y * width + x] === color;
  const putPixel = (x, y, color) => data[y * width + x] = color;
  const comboPush = (xLeft, xRight, checkY) => {
    let checkX = xLeft + 1;
    let isCombo = !1;
    for (;checkX < xRight; ) {
      if (checkPixelColor(checkX, checkY, fromColor)) {
        if (!isCombo) {
          isCombo = !0;
          markPointList.push({
            x: checkX,
            y: checkY
          });
        }
      } else isCombo = !1;
      checkX++;
    }
  };
  for (;markPointList.length; ) {
    const {x, y} = markPointList.pop();
    if (!checkPixelColor(x, y, fromColor)) continue;
    putPixel(x, y, toColor);
    let xLeft = x - 1;
    for (;xLeft >= 0 && checkPixelColor(xLeft, y, fromColor); ) {
      putPixel(xLeft, y, toColor);
      xLeft--;
    }
    let xRight = x + 1;
    for (;xRight < width && checkPixelColor(xRight, y, fromColor); ) {
      putPixel(xRight, y, toColor);
      xRight++;
    }
    y - 1 >= 0 && comboPush(xLeft, xRight, y - 1);
    y + 1 < height && comboPush(xLeft, xRight, y + 1);
  }
};
export { getPixelColor, replacePixelColor, drawPixel, drawPixelLine, drawPixelLineList, scale, crop, floodFill };