import { clock } from "../../common/time.js";
import { getDist } from "../../common/geometry/D2/Vector.js";
import { isContainPoint as isBoundingRectContainPoint } from "../../common/geometry/D2/BoundingRect.js";
const {document, getComputedStyle, TouchEvent} = window;
const POINTER_EVENT_TYPE = {
  START: "START",
  MOVE: "MOVE",
  END: "END",
  CANCEL: "CANCEL"
};
const applyPointerEventListener = ({element, onEvent, isGlobal = !1, isCancel = !0, isCancelOnOutOfBound = !0, isUseTouchEvent = !0}) => {
  !getComputedStyle(element).touchAction && console.warn('[applyPointerEventListener] should set CSS "touch-action" to "none" to prevent browser defaults');
  const eventSource = isGlobal ? document : element;
  isCancelOnOutOfBound = isCancel && isCancelOnOutOfBound;
  let timeStart;
  let pointStart;
  let eventStart;
  let prevEvent;
  let prevState;
  const reset = () => {
    timeStart = null;
    pointStart = null;
    eventStart = null;
    prevEvent = null;
    prevState = null;
    eventSource.removeEventListener("pointermove", onMove);
    eventSource.removeEventListener("pointerup", onEnd);
    isCancel && eventSource.removeEventListener("pointercancel", onCancel);
    isCancelOnOutOfBound && eventSource.removeEventListener("pointerleave", onCancel);
    isUseTouchEvent && eventSource.removeEventListener("touchmove", onMove, {
      passive: !1
    });
    isUseTouchEvent && eventSource.removeEventListener("touchend", onEnd);
    isUseTouchEvent && isCancel && eventSource.removeEventListener("touchcancel", onCancel);
  };
  const getEventPoint = isUseTouchEvent && TouchEvent ? event => event instanceof TouchEvent ? event.touches.length ? {
    x: event.touches[0].clientX,
    y: event.touches[0].clientY
  } : {
    x: event.changedTouches[0].clientX,
    y: event.changedTouches[0].clientY
  } : {
    x: event.clientX,
    y: event.clientY
  } : event => ({
    x: event.clientX,
    y: event.clientY
  });
  const calcState = event => {
    if (event !== prevEvent) {
      const time = clock();
      const point = getEventPoint(event);
      prevEvent = event;
      prevState = {
        timeStart,
        pointStart,
        eventStart,
        time,
        point,
        event,
        duration: time - timeStart,
        distance: getDist(pointStart, point)
      };
    }
    return prevState;
  };
  const checkShouldPass = isUseTouchEvent && TouchEvent ? event => event instanceof TouchEvent ? "touchend" === event.type || "touchcancel" === event.type ? 0 !== event.touches.length : 1 !== event.touches.length : !event.isPrimary || "touch" === event.pointerType : event => !event.isPrimary;
  const onStart = event => {
    if (!checkShouldPass(event)) {
      timeStart = clock();
      pointStart = getEventPoint(event);
      eventStart = event;
      eventSource.addEventListener("pointermove", onMove);
      eventSource.addEventListener("pointerup", onEnd);
      isCancel && eventSource.addEventListener("pointercancel", onCancel);
      isCancelOnOutOfBound && eventSource.addEventListener("pointerleave", onCancel);
      isUseTouchEvent && eventSource.addEventListener("touchmove", onMove, {
        passive: !1
      });
      isUseTouchEvent && eventSource.addEventListener("touchend", onEnd);
      isUseTouchEvent && isCancel && eventSource.addEventListener("touchcancel", onCancel);
      onEvent(POINTER_EVENT_TYPE.START, event, calcState);
    }
  };
  const onMove = event => {
    if (checkShouldPass(event)) return;
    const {point} = calcState(event);
    if (isCancelOnOutOfBound && !isBoundingRectContainPoint(element.getBoundingClientRect(), point)) return onCancel(event);
    onEvent(POINTER_EVENT_TYPE.MOVE, event, calcState);
  };
  const onEnd = event => {
    if (!checkShouldPass(event)) {
      onEvent(POINTER_EVENT_TYPE.END, event, calcState);
      reset();
    }
  };
  const onCancel = event => {
    if (!checkShouldPass(event)) {
      onEvent(POINTER_EVENT_TYPE.CANCEL, event, calcState);
      reset();
    }
  };
  element.addEventListener("pointerdown", onStart);
  isUseTouchEvent && element.addEventListener("touchstart", onStart, {
    passive: !1
  });
  reset();
  return () => {
    element.removeEventListener("pointerdown", onStart);
    isUseTouchEvent && element.removeEventListener("touchstart", onStart, {
      passive: !1
    });
    reset();
  };
};
const ENHANCED_POINTER_EVENT_TYPE = {
  TAP: "TAP",
  HOLD: "HOLD",
  DRAG_MOVE: "DRAG_MOVE",
  DRAG_END: "DRAG_END",
  DRAG_CANCEL: "DRAG_CANCEL"
};
const applyEnhancedPointerEventListener = ({onEnhancedEvent, onEvent, holdDurationThreshold = 500, dragDistanceThreshold = 5, ...extraOption}) => {
  let isDragging = !1;
  return applyPointerEventListener({
    onEvent: (name, event, calcState) => {
      const eventState = calcState(event);
      switch (name) {
       case POINTER_EVENT_TYPE.START:
        isDragging = !1;
        break;
       case POINTER_EVENT_TYPE.MOVE:
        isDragging |= eventState.distance >= dragDistanceThreshold;
        isDragging && onEnhancedEvent(ENHANCED_POINTER_EVENT_TYPE.DRAG_MOVE, eventState);
        break;
       case POINTER_EVENT_TYPE.END:
        isDragging ? onEnhancedEvent(ENHANCED_POINTER_EVENT_TYPE.DRAG_END, eventState) : eventState.duration >= holdDurationThreshold ? onEnhancedEvent(ENHANCED_POINTER_EVENT_TYPE.HOLD, eventState) : onEnhancedEvent(ENHANCED_POINTER_EVENT_TYPE.TAP, eventState);
        break;
       case POINTER_EVENT_TYPE.CANCEL:
        isDragging && onEnhancedEvent(ENHANCED_POINTER_EVENT_TYPE.DRAG_CANCEL, eventState);
      }
      onEvent && onEvent(name, event, calcState);
    },
    ...extraOption
  });
};
export { POINTER_EVENT_TYPE, applyPointerEventListener, ENHANCED_POINTER_EVENT_TYPE, applyEnhancedPointerEventListener };