import { remessageError } from "../common/error.js";
const {XMLHttpRequest, Blob, TextDecoder} = window;
const REGEXP_HEADER_SEPARATOR = /[\r\n]+/;
const fetchLikeRequest = (url, {method = "GET", headers: requestHeaders, body, timeout = 0, credentials, onProgressUpload, onProgressDownload} = {}) => new Promise(((resolve, reject) => {
  const request = new XMLHttpRequest;
  const tagError = error => remessageError(Object.assign(error, {
    url,
    method
  }), `[${method}|${url}] ${error.message}`);
  request.onerror = () => reject(tagError(new Error("NETWORK_ERROR")));
  request.ontimeout = () => reject(tagError(new Error("NETWORK_TIMEOUT")));
  request.onreadystatechange = () => {
    const {readyState, status} = request;
    if (2 !== readyState || 0 === status) return;
    const responseHeaders = request.getAllResponseHeaders().split(REGEXP_HEADER_SEPARATOR).reduce(((o, rawHeader) => {
      const [key, ...valueList] = rawHeader.split(":");
      valueList.length && (o[key.trim().toLowerCase()] = valueList.join(":").trim());
      return o;
    }), {});
    resolve({
      status,
      ok: status >= 200 && status < 300,
      headers: responseHeaders,
      ..._wrapPayload(request, tagError)
    });
  };
  onProgressUpload && request.upload && (request.upload.onprogress = _wrapOnProgress(onProgressUpload));
  onProgressDownload && (request.onprogress = _wrapOnProgress(onProgressDownload));
  request.open(method, url);
  requestHeaders && Object.entries(requestHeaders).forEach((([key, value]) => request.setRequestHeader(key, value)));
  request.responseType = "arraybuffer";
  request.timeout = timeout || 0;
  request.withCredentials = "include" === credentials;
  request.send(body || null);
}));
const _wrapOnProgress = onProgress => ({lengthComputable, loaded, total}) => {
  onProgress(loaded, lengthComputable ? total : 1 / 0);
};
const _wrapPayload = (request, tagError) => {
  let payloadOutcome;
  setTimeout((() => {
    if (!payloadOutcome) {
      payloadOutcome = "DROP";
      request.abort();
    }
  }));
  const _arrayBuffer = () => new Promise(((resolve, reject) => {
    if (payloadOutcome) return reject(new Error("KEEP" === payloadOutcome ? "PAYLOAD_ALREADY_USED" : "PAYLOAD_ALREADY_DROPPED"));
    payloadOutcome = "KEEP";
    request.onload = () => resolve(request.response);
    request.onerror = () => reject(new Error("PAYLOAD_ERROR"));
    request.ontimeout = () => reject(new Error("PAYLOAD_TIMEOUT"));
  }));
  const _onReject = error => {
    throw tagError(error);
  };
  return {
    arrayBuffer: () => _arrayBuffer().catch(_onReject),
    blob: () => _arrayBuffer().then((arrayBuffer => new Blob([ arrayBuffer ]))).catch(_onReject),
    text: () => _arrayBuffer().then((arrayBuffer => (new TextDecoder).decode(arrayBuffer))).catch(_onReject),
    json: () => _arrayBuffer().then((arrayBuffer => JSON.parse((new TextDecoder).decode(arrayBuffer)))).catch(_onReject)
  };
};
export { fetchLikeRequest };