import { createHub } from "../module/Event.js";
import { createDoublyLinkedList, createNode } from "./LinkedList.js";
/** @typedef { { key: vJSON, size: number, expireAt: number } & LinkedListNode } CacheMapCache */
/** @type { (key: vJSON, value: vJSON, size: number, expireAt: number) => CacheMapCache } */ const createCache = (key, value, size, expireAt) => ({
  ...createNode(value),
  key,
  size,
  expireAt
}
/** @typedef { (key: vJSON, value: vJSON, size?: number, expireAt?: number) => void } SetCacheMap */
/** @typedef { (key: vJSON, time: number) => vJSON | void } GetCacheMap */
/** @typedef { (key: vJSON) => vJSON } DeleteCacheMap */
/** @typedef { { value: vJSON, key: vJSON, size: number, expireAt: number } } CacheMapCacheSave */
/** @typedef { {
 * hasEventHub: boolean, clearEventHub: EventHub['clear'], subscribe: EventHub['subscribe'], unsubscribe: EventHub['unsubscribe'],
 * clear: GetVoid,
 * getSize: GetNumber, getValueSizeSum: GetNumber,
 * set: SetCacheMap, get: GetCacheMap, touch: GetCacheMap, delete: DeleteCacheMap,
 * saveCacheList: () => CacheMapCacheSave[], loadCacheList: (v: CacheMapCacheSave[], t?: number) => void,
 * } } CacheMap */
/** @type { (option: {
 *   valueSizeSumMax: number,
 *   valueSizeSingleMax?: number,
 *   expireAfter?: number,
 *   eventHub?: EventHub | null
 * }) => CacheMap } */);
const createCacheMap = ({valueSizeSumMax, valueSizeSingleMax = Math.max(.05 * valueSizeSumMax, 1), expireAfter = 6e4, eventHub = createHub()}) => {
  const hasEventHub = Boolean(eventHub);
  const {clear: clearEventHub, subscribe, unsubscribe, send} = /** @type { ?EventHub } */ hasEventHub ? eventHub : {};
  /** @type { Map<vJSON, CacheMapCache> } */  const map = new Map;
  const linkedList = createDoublyLinkedList();
  let valueSizeSum = 0;
  /** @type { (CacheMapCache) => void } */  const cacheAdd = cache => {
    map.set(cache.key, cache);
    linkedList.unshift(cache);
    valueSizeSum += cache.size;
    hasEventHub && send({
      type: "add",
      key: cache.key,
      payload: cache.value
    });
  };
  /** @type { (CacheMapCache) => void } */  const cacheDelete = cache => {
    map.delete(cache.key);
    linkedList.remove(cache);
    valueSizeSum -= cache.size;
    hasEventHub && send({
      type: "delete",
      key: cache.key,
      payload: cache.value
    });
  };
  /** @type { CacheMap } */  return {
    hasEventHub,
    clearEventHub,
    subscribe,
    unsubscribe,
    clear: () => map.forEach(cacheDelete),
    getSize: linkedList.getLength,
    getValueSizeSum: () => valueSizeSum
    /** @type { SetCacheMap } */ ,
    set: (key, value, size = 1, expireAt = Date.now() + expireAfter) => {
      const prevCache = map.get(key);
      prevCache && cacheDelete(prevCache);
      if (!(size > valueSizeSingleMax)) {
        for (;size + valueSizeSum > valueSizeSumMax; ) cacheDelete(linkedList.getTail().prev);
        cacheAdd(createCache(key, value, size, expireAt));
      }
    },
    /** @type { GetCacheMap } */
    get: (key, time = Date.now()) => {
      const cache = map.get(key);
      if (cache) {
        if (cache.expireAt <= time) return cacheDelete(cache);
        linkedList.moveToFirst(cache);
        return cache.value;
      }
    },
    /** @type { GetCacheMap } */
    touch: (key, expireAt = Date.now() + expireAfter) => {
      const cache = map.get(key);
      if (cache) {
        cache.expireAt = expireAt;
        linkedList.moveToFirst(cache);
        return cache.value;
      }
    },
    delete: key => {
      const cache = map.get(key);
      cache && cacheDelete(cache);
      return cache && cache.value;
    },
    saveCacheList: () => {
      /** @type { CacheMapCacheSave[] } */
      const cacheList = [];
      linkedList.forEachReverse((/** @type { CacheMapCache } */ node => {
        const {key, value, size, expireAt} = node;
        cacheList.push({
          key,
          value,
          size,
          expireAt
        });
      }));
      return cacheList;
    },
    loadCacheList: (cacheList, time = Date.now()) => cacheList.forEach((({key, value, size, expireAt}) => {
      if (expireAt <= time) return;
      const cache = map.get(key);
      if (!(cache && cache.expireAt >= expireAt)) {
        cache && cacheDelete(cache);
        cacheAdd(createCache(key, value, size, expireAt));
      }
    }))
  };
};
export { createCache, createCacheMap };