/** @typedef { { value: vJSON, prev: LinkedListNode | null, next: LinkedListNode | null } } LinkedListNode */
/** @type { (value: vJSON, prev?: LinkedListNode, next?: LinkedListNode) => LinkedListNode } */
const createNode = (value, prev = null, next = null) => ({
  value,
  prev,
  next
})
/** @typedef { () => LinkedListNode } GetLinkedListNode */
/** @typedef { (v: LinkedListNode) => void } SetLinkedListNode */
/** @typedef { (a: LinkedListNode, b: LinkedListNode) => void } Set2LinkedListNode */
/** @typedef { (callback: (v: LinkedListNode, i: number) => void) => void } ForEachLinkedListNode */
/** @typedef { {
 *   clear: GetVoid,
 *   getHead: GetLinkedListNode, getTail: GetLinkedListNode,
 *   getLength: GetNumber,
 *   insertAfter: Set2LinkedListNode, insertBefore: Set2LinkedListNode,
 *   remove: SetLinkedListNode, removeBetween: Set2LinkedListNode,
 *   forEach: ForEachLinkedListNode, forEachReverse: ForEachLinkedListNode,
 *   reverse: GetVoid,
 *   moveToFirst: SetLinkedListNode, moveToLast: SetLinkedListNode,
 *   push: SetLinkedListNode, pop: GetVoid,
 *   unshift: SetLinkedListNode, shift: GetVoid
 * } } DoublyLinkedList */
/** @type { () => DoublyLinkedList } */;
const createDoublyLinkedList = () => {
  /** @type { LinkedListNode } */
  let head;
  /** @type { LinkedListNode } */  let tail;
  /** @type { number } */  let length;
  /** @type { GetVoid } */  const clear = () => {
    head = createNode(null);
    tail = createNode(null, head);
    head.next = tail;
    length = 0;
  };
  clear();
  /** @type { Set2LinkedListNode } */  const insertAfter = (node, prevNode) => {
    const {next} = prevNode;
    node.prev = prevNode;
    node.next = next;
    prevNode.next = next.prev = node;
    length++;
  };
  /** @type { Set2LinkedListNode } */  const insertBefore = (node, nextNode) => {
    const {prev} = nextNode;
    node.prev = prev;
    node.next = nextNode;
    nextNode.prev = prev.next = node;
    length++;
  };
  /** @type { SetLinkedListNode } */  const remove = node => {
    const {prev, next} = node;
    prev.next = next;
    next.prev = prev;
    node.prev = node.next = null;
    length--;
  };
  /** @type { Set2LinkedListNode } */  return {
    clear,
    getHead: () => head,
    getTail: () => tail,
    getLength: () => length,
    insertAfter,
    insertBefore,
    remove,
    removeBetween: (fromNode, toNode) => {
      const {prev} = fromNode;
      const {next} = toNode;
      prev.next = next;
      next.prev = prev;
      fromNode.prev = toNode.next = null;
      let node = fromNode;
      for (;node; ) {
        length--;
        node = node.next;
      }
    },
    forEach: callback => {
      let node = head.next;
      let index = 0;
      for (;node !== tail; ) {
        callback(node, index);
        node = node.next;
        index++;
      }
    },
    forEachReverse: callback => {
      let node = tail.prev;
      let index = length - 1;
      for (;node !== head; ) {
        callback(node, index);
        node = node.prev;
        index--;
      }
    },
    reverse: () => {
      let node = head.next;
      for (;node !== tail; ) {
        const {prev, next} = node;
        node.prev = next;
        node.next = prev;
        node = next;
      }
      const {next} = head;
      const {prev} = tail;
      head.next = prev;
      tail.prev = next;
      prev.prev = head;
      next.next = tail;
    },
    moveToFirst: node => {
      if (node === head.next) return;
      const {prev, next} = node;
      prev.next = next;
      next.prev = prev;
      node.prev = head;
      node.next = head.next;
      node.next.prev = head.next = node;
    },
    moveToLast: node => {
      if (node === tail.prev) return;
      const {prev, next} = node;
      prev.next = next;
      next.prev = prev;
      node.next = tail;
      node.prev = tail.prev;
      node.prev.next = tail.prev = node;
    },
    push: node => insertBefore(node, tail),
    pop: () => remove(tail.prev),
    unshift: node => insertAfter(node, head),
    shift: () => remove(head.next)
  };
};
export { createNode, createDoublyLinkedList };