import { clock, setWeakTimeout, setTimeoutAsync } from "./time.js";
import { rethrowError } from "./error.js";
import { isPromiseAlike } from "./check.js";
/** @deprecated */ const debounce = (func, wait = 250, isLeadingEdge = !1) => (isLeadingEdge ? debounceL : debounceT)(func, wait);
const debounceT = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    clearTimeout(timeoutToken);
    timeoutToken = setTimeout((() => {
      timeoutToken = null;
      func.apply(null, args);
    }), wait);
  };
};
const debounceL = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    const isCallNow = null === timeoutToken;
    clearTimeout(timeoutToken);
    timeoutToken = setTimeout((() => {
      timeoutToken = null;
    }), wait);
    isCallNow && func.apply(null, args);
  };
};
/** @deprecated */ const throttle = (func, wait = 250, isLeadingEdge = !1) => (isLeadingEdge ? throttleL : throttleTE)(func, wait)
/** @deprecated */ const throttleTE = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    timeoutToken || (timeoutToken = setWeakTimeout((() => {
      timeoutToken = null;
      func.apply(null, args);
    }), wait));
  };
};
const throttleT = (func, wait = 250) => {
  let timeoutToken = null;
  let lastArgs = null;
  return (...args) => {
    lastArgs = args;
    timeoutToken || (timeoutToken = setWeakTimeout((() => {
      timeoutToken = null;
      func.apply(null, lastArgs);
      lastArgs = null;
    }), wait));
  };
};
const throttleL = (func, wait = 250) => {
  let timeoutToken = null;
  return (...args) => {
    if (timeoutToken) return;
    const isCallNow = null === timeoutToken;
    timeoutToken = setWeakTimeout((() => {
      timeoutToken = null;
    }), wait);
    isCallNow && func.apply(null, args);
  };
};
const once = func => {
  let isCalled = !1;
  return (...args) => {
    if (!0 !== isCalled) {
      isCalled = !0;
      return func.apply(null, args);
    }
  };
};
/** @type { (asyncFunc: Function, onError?: Function) => {
  trigger: Function,
  getRunningPromise: () => Promise | undefined
} } */ const lossyAsync = (asyncFunc, onError = rethrowError) => {
  let runningPromise;
  const onResolve = () => {
    runningPromise = void 0;
  };
  const onReject = error => {
    runningPromise = void 0;
    onError(error);
  };
  return {
    trigger: (...args) => {
      if (!runningPromise) try {
        const result = asyncFunc.apply(null, args);
        isPromiseAlike(result) ? runningPromise = result.then(onResolve, onReject) : onResolve();
      } catch (error) {
        onReject(error);
      }
    },
    getRunningPromise: () => runningPromise
  };
};
const loneAsync = asyncFunc => {
  let runningPromise;
  const onEnd = () => {
    runningPromise = void 0;
  };
  return (...args) => {
    if (runningPromise) return runningPromise;
    const result = asyncFunc.apply(null, args);
    if (isPromiseAlike(result)) {
      result.then(onEnd, onEnd);
      runningPromise = result;
    }
    return result;
  };
};
const withCache = func => {
  let cache;
  return () => {
    void 0 === cache && (cache = func());
    return cache;
  };
};
const withCacheAsync = asyncFunc => {
  let cache;
  return async () => {
    void 0 === cache && (cache = await asyncFunc());
    return cache;
  };
};
/** @deprecated */ const withDelayArgvQueue = (func, delayWrapper = debounce, ...args) => {
  let argvQueue = [];
  const delayFunc = delayWrapper((() => {
    const currentArgvQueue = argvQueue;
    argvQueue = [];
    func(currentArgvQueue);
  }), ...args);
  return (...argv) => {
    argvQueue.push(argv);
    delayFunc();
  };
};
const withRepeat = (func, count = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    func(looped, count);
    looped++;
  }
};
const withRepeatAsync = async (func, count = 0, wait = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    const startTime = clock();
    await func(looped, count);
    looped++;
    const remainingTime = wait - (clock() - startTime);
    remainingTime > 0 && await setTimeoutAsync(remainingTime);
  }
};
const withRetry = (func, maxRetry = 1 / 0) => {
  let failed = 0;
  for (;;) try {
    return func(failed, maxRetry);
  } catch (error) {
    failed++;
    if (maxRetry < failed) throw error;
  }
};
const withRetryAsync = async (func, maxRetry = 1 / 0, wait = 0) => {
  let failed = 0;
  for (;;) {
    const startTime = clock();
    try {
      return await func(failed, maxRetry);
    } catch (error) {
      failed++;
      if (maxRetry < failed) throw error;
      const remainingTime = wait - (clock() - startTime);
      remainingTime > 0 && await setTimeoutAsync(remainingTime);
    }
  }
};
const withTimeoutAsync = (func, timeout, message = "") => withTimeoutPromise(func(), timeout, message);
const withTimeoutPromise = (promise, timeout, message = "") => {
  if (!(timeout > 0)) return promise;
  let timeoutToken = null;
  return Promise.race([ promise, new Promise(((resolve, reject) => {
    timeoutToken = setWeakTimeout((() => reject(DUMMY_ERROR)), timeout);
  })) ]).then((result => {
    clearTimeout(timeoutToken);
    return result;
  }), (error => {
    clearTimeout(timeoutToken);
    throw error !== DUMMY_ERROR ? error : new Error(`timeout after: ${timeout}${message ? `, ${message}` : ""}`);
  }));
};
const DUMMY_ERROR = {};
/** @type { <T>() => { promise: Promise<T>, resolve: (v?: T) => void, reject: (e: Error) => void } } */ const createInsideOutPromise = () => {
  let promiseResolve, promiseReject;
  return {
    promise: new Promise(((resolve, reject) => {
      promiseResolve = resolve;
      promiseReject = reject;
    })),
    resolve: promiseResolve,
    reject: promiseReject
  };
};
const runAsPromise = func => {
  try {
    const result = func();
    return isPromiseAlike(result) ? result : Promise.resolve(result);
  } catch (error) {
    return Promise.reject(error);
  }
};
const runAsyncByLane = (laneSize, asyncFuncList = []) => {
  if (0 === (laneSize = Math.min(laneSize, asyncFuncList.length))) return Promise.resolve();
  const afList = [ ...asyncFuncList ];
  const iop = createInsideOutPromise();
  let isAbort = !1;
  let activeTask = 0;
  const onResolve = () => {
    if (!isAbort) {
      activeTask--;
      afList.length || activeTask ? runMore() : iop.resolve();
    }
  };
  const onReject = error => {
    isAbort = !0;
    iop.reject(error);
  };
  const runMore = () => {
    if (!isAbort) for (;activeTask < laneSize && afList.length; ) {
      activeTask++;
      runAsPromise(afList.shift()).then(onResolve, onReject);
    }
  };
  runMore();
  return iop.promise;
};
export { debounce, debounceT, debounceL, throttle, throttleT, throttleL, once, lossyAsync, loneAsync, withCache, withCacheAsync, withDelayArgvQueue, withRepeat, withRepeatAsync, withRetry, withRetryAsync, withTimeoutAsync, withTimeoutPromise, createInsideOutPromise, runAsPromise, runAsyncByLane };