import { getSample } from "../math/sample.js";
import { createAsyncFuncQueue } from "./AsyncFuncQueue.js";
const createAsyncLane = ({laneSize, createAsyncQueue = createAsyncFuncQueue}) => {
  if (!((laneSize = Number(laneSize)) >= 1)) throw new Error(`invalid laneSize: ${laneSize}`);
  const laneList = getSample((index => ({
    index,
    asyncQueue: createAsyncQueue()
  })), laneSize);
  return {
    laneList,
    calcStatus: (func = DEFAULT_STATUS_MAP_FUNC) => laneList.map(func),
    getTailPromise: () => Promise.all(laneList.map((({asyncQueue: {getTailPromise}}) => getTailPromise()))),
    reset: () => laneList.forEach((({asyncQueue}) => {
      asyncQueue.reset();
    })),
    push: (value, laneIndex) => {
      const lane = laneList[laneIndex];
      if (!lane) throw new Error(`invalid laneIndex: ${laneIndex}`);
      return lane.asyncQueue.push(value);
    }
  };
};
const DEFAULT_STATUS_MAP_FUNC = ({index, asyncQueue}) => ({
  index,
  length: asyncQueue.getLength()
});
const extendAutoSelectLane = (asyncLane, selectLane = selectMinLoadLane) => {
  const {laneList} = asyncLane;
  return {
    ...asyncLane,
    pushAuto: value => asyncLane.push(value, selectLane(laneList, value).index)
  };
};
const selectMinLoadLane = (laneList, value) => laneList.reduce(((o, lane) => o.asyncQueue.getLength() > lane.asyncQueue.getLength() ? lane : o), laneList[0]);
const extendLaneValueList = asyncLane => {
  const {laneList} = asyncLane;
  laneList.forEach((lane => {
    lane.valueList = [];
  }));
  return {
    ...asyncLane,
    calcStatus: func => {
      const status = asyncLane.calcStatus(func);
      void 0 === func && status.forEach((laneStatus => {
        laneStatus.valueList = laneList[laneStatus.index].valueList;
      }));
      return status;
    },
    reset: () => {
      asyncLane.reset();
      laneList.forEach((({valueList}) => {
        valueList.length = 0;
      }));
    },
    push: (value, laneIndex) => {
      const valuePromise = asyncLane.push(value, laneIndex);
      const {asyncQueue, valueList} = laneList[laneIndex];
      valueList.unshift(value);
      valueList.length = asyncQueue.getLength();
      return valuePromise;
    },
    trimValueList: () => laneList.forEach((({asyncQueue, valueList}) => {
      valueList.length = asyncQueue.getLength();
    }))
  };
};
const extendLaneValueMap = asyncLane => {
  const {laneList} = asyncLane;
  laneList.forEach((lane => {
    lane.valueMap = new Map;
  }));
  return {
    ...asyncLane,
    calcStatus: func => {
      const status = asyncLane.calcStatus(func);
      void 0 === func && status.forEach((laneStatus => {
        laneStatus.valueList = [ ...laneList[laneStatus.index].valueMap.values() ];
      }));
      return status;
    },
    reset: () => {
      asyncLane.reset();
      laneList.forEach((({valueMap}) => {
        valueMap.clear();
      }));
    },
    push: (value, laneIndex) => {
      if (!value.id) throw new Error("expect value.id");
      const valuePromise = asyncLane.push(value, laneIndex);
      laneList[laneIndex].valueMap.set(value.id, value);
      return valuePromise;
    },
    trimValueMap: filterFunc => laneList.forEach((({valueMap}) => valueMap.forEach(((value, id) => filterFunc(value, id) || valueMap.delete(id))))),
    findValue: (id, laneIndex) => {
      const lane = laneList[laneIndex];
      return lane && lane.valueMap.get(id);
    },
    dropValue: (id, laneIndex) => {
      const lane = laneList[laneIndex];
      return lane && lane.valueMap.delete(id);
    }
  };
};
const extendAutoSelectByTagLane = (asyncLane, selectByTagLane = selectByTagOrMinLoadLane) => {
  const {laneList} = asyncLane;
  laneList.forEach((lane => {
    lane.tagList = [];
  }));
  const push = (value, laneIndex, tag) => {
    const valuePromise = asyncLane.push(value, laneIndex);
    const {asyncQueue, tagList} = laneList[laneIndex];
    tagList.unshift(tag);
    tagList.length = asyncQueue.getLength();
    return valuePromise;
  };
  return {
    ...asyncLane,
    calcStatus: func => {
      const status = asyncLane.calcStatus(func);
      void 0 === func && status.forEach((laneStatus => {
        laneStatus.tagList = laneList[laneStatus.index].tagList;
      }));
      return status;
    },
    reset: () => {
      asyncLane.reset();
      laneList.forEach((({tagList}) => {
        tagList.length = 0;
      }));
    },
    push,
    pushAutoTag: (value, tag) => push(value, selectByTagLane(laneList, value, tag).index, tag)
  };
};
const selectByTagOrMinLoadLane = (laneList, value, tag) => tag && laneList.find((lane => lane.tagList.includes(tag))) || selectMinLoadLane(laneList);
export { createAsyncLane, extendAutoSelectLane, selectMinLoadLane, extendLaneValueList, extendLaneValueMap, extendAutoSelectByTagLane, selectByTagOrMinLoadLane };