/** @typedef { (data: any) => void } EHListener */
/** @typedef { {
 clear: () => void,
 subscribe: (listener: EHListener) => void,
 unsubscribe: (listener: EHListener) => void,
 send: (data: any) => void
 } } EventHub */
/** @type { () => EventHub } */
const createHub = () => {
  /** @type { Set<EHListener> } */
  const set = new Set;
  /** @type { EventHub['clear'] } */  return {
    clear: () => set.clear()
    /** @type { EventHub['subscribe'] } */ ,
    subscribe: listener => {
      set.add(listener);
    },
    unsubscribe: listener => {
      set.delete(listener);
    },
    send: data => set.forEach((listener => listener(data)))
  };
};
/** @typedef { (...args: any[]) => void } EBListener */
/** @typedef { {
 setMap: Map<string, Set<EBListener>>,
 clear: () => void,
 on: (type: string, listener: EBListener) => void,
 off: (type: string, listener: EBListener) => void,
 } } EventBase - base type for EventTargetAlike & EventEmitterAlike */
/** @type { () => EventBase } */ const createEventBase = () => {
  const setMap = new Map;
  return {
    setMap,
    clear: () => setMap.clear()
    /** @type { EventBase['on'] } */ ,
    on: (type, listener) => {
      const listenerSet = setMap.get(type);
      listenerSet ? listenerSet.add(listener) : setMap.set(type, new Set([ listener ]));
    },
    off: (type, listener) => {
      const listenerSet = setMap.get(type);
      listenerSet && listenerSet.delete(listener);
      listenerSet && 0 === listenerSet.size && setMap.delete(type);
    }
  };
};
/** @typedef { { type: string } } ETAEvent */
/** @typedef { (event: ETAEvent) => void } ETAListener */
/** @typedef { {
 clear: () => void,
 dispatchEvent: (v: ETAEvent) => void,
 addEventListener: (type: string, listener: ETAListener) => void,
 removeEventListener: (type: string, listener: ETAListener) => void
 } } EventTargetAlike */
/** @type { () => EventTargetAlike } */ const createEventTarget = () => {
  const {setMap, clear, on, off} = createEventBase();
  /** @type { EventTargetAlike['dispatchEvent'] } */  return {
    clear,
    dispatchEvent: event => {
      /** @type { Set<ETAListener> } */
      const listenerSet = setMap.get(event.type);
      listenerSet && listenerSet.forEach((listener => listener(event)));
    },
    addEventListener: on,
    removeEventListener: off
  };
};
/** @typedef { (...args: any[]) => void } EEAListener */
/** @typedef { {
 *   clear: () => void,
 *   emit: EEAListener,
 *   on: (type: string, listener: EEAListener) => void,
 *   off: (type: string, listener: EEAListener) => void,
 *   addListener: (type: string, listener: EEAListener) => void,
 *   removeListener: (type: string, listener: EEAListener) => void,
 *   removeAllListeners: (...eventNameList: string[]) => void
 * } } EventEmitterAlike */
/** @type { () => EventEmitterAlike } */ const createEventEmitter = () => {
  const {setMap, clear, on, off} = createEventBase();
  /** @type { EventEmitterAlike['removeAllListeners'] } */  return {
    clear,
    emit: (eventName, ...args) => {
      /** @type { Set<EEAListener> } */
      const listenerSet = setMap.get(eventName);
      listenerSet && listenerSet.forEach((listener => listener(...args)));
    },
    on,
    off,
    addListener: on,
    removeListener: off,
    removeAllListeners: (...eventNameList) => {
      eventNameList.length ? eventNameList.forEach((eventName => setMap.delete(eventName))) : setMap.clear();
    }
  };
};
export { createHub, createEventTarget, createEventEmitter };