import { isString, isObjectAlike, isBasicFunction } from "../check.js";
import { getRandomId62S } from "../math/random.js";
/** @typedef { Error & { exotId?: string } } ExotError */
/** @type { (exotId: string, message: string) => ExotError } */ const createExotError = (exotId, message) => {
  /** @type { ExotError } */
  const exotError = new Error(message);
  exotError.exotId = exotId;
  return exotError;
};
/** @typedef { (error: ExotError) => void } OnExotError */
/** @typedef { {
 id: string,
 up: (onExotError: OnExotError) => Promise<void> | void,
 down: () => Promise<void> | void,
 isUp: () => boolean
 } } Exot */
/** @type { (opt?: { idPrefix?: string, id?: string, onUp?: Exot['up'], onDown?: Exot['down'] }) => Exot } */ const createDummyExot = ({idPrefix = "DUMMY-EXOT-", id = getRandomId62S(idPrefix), onUp, onDown, ...extra} = {}) => {
  let isActive = !1;
  return {
    ...extra,
    id,
    up: async onExotError => {
      if (isActive) throw new Error("already up");
      if (!onExotError) throw new Error("expect onExotError to receive ExotError notice");
      onUp && await onUp(onExotError);
      isActive = !0;
    },
    down: async () => {
      if (isActive) {
        isActive = !1;
        onDown && await onDown();
      }
    },
    isUp: () => isActive
  };
};
/** @typedef { Exot & {
 upEach: Exot['up'], upBatch: Exot['up'],
 downEach: Exot['down'], downBatch: Exot['down'],
 getSize: () => number, set: (exot: Exot) => void, get: (exotId: string) => Exot | undefined, delete: (exotId: string) => Exot | undefined,
 load: (exot: Exot, onExotError: OnExotError) => Promise<void>, drop: (exotId: string) => Promise<Exot>,
 exotMap: ExotMap
 } } ExotGroup */
/** @type { (opt: {
 idPrefix?: string, id?: string,
 getOnExotError?: (exotGroup: ExotGroup) => OnExotError, onExotError?: OnExotError,
 exotList?: Exot[], exotMap?: ExotMap,
 isBatch?: boolean
 }) => ExotGroup } */ const createExotGroup = ({idPrefix = "EXOT-GROUP-", id = getRandomId62S(idPrefix), getOnExotError, onExotError, exotList = [], exotMap = toExotMap(...exotList), isBatch = !1} = {}) => {
  const upEach = async (onExotError = onExotErrorGroup) => {
    for (const exot of exotMap.values()) exot.isUp() || await exot.up(onExotError);
  };
  const upBatch = async (onExotError = onExotErrorGroup) => {
    await Promise.all(mapExotMapValue(exotMap, (exot => exot.isUp() || exot.up(onExotError))));
  };
  const downEach = async () => {
    for (const exot of Array.from(exotMap.values()).reverse()) await exot.down();
  };
  const downBatch = async () => {
    await Promise.all(Array.from(exotMap.values()).reverse().map((exot => exot.down())));
  };
  const set = exot => {
    if (!isExot(exot)) throw new Error(`invalid exot: ${exot}`);
    if (exotMap.has(exot.id)) throw new Error(`duplicate exot id: ${exot.id}`);
    exotMap.set(exot.id, exot);
  };
  const get = exotId => exotMap.get(exotId);
  const deleteKeyword = exotId => {
    const exot = get(exotId);
    void 0 !== exot && exotMap.delete(exotId);
    return exot;
  };
  const down = isBatch ? downBatch : downEach;
  const exotGroup = {
    id,
    up: isBatch ? upBatch : upEach,
    down,
    isUp: () => findExotMapValue(exotMap, (exot => exot.isUp())),
    upEach,
    upBatch,
    downEach,
    downBatch,
    getSize: () => exotMap.size,
    set,
    get,
    delete: deleteKeyword,
    load: async (exot, onExotError = onExotErrorGroup) => {
      set(exot);
      exot.isUp() || await exot.up(onExotError);
    },
    drop: async exotId => {
      const exot = deleteKeyword(exotId);
      exot && await exot.down();
      return exot;
    },
    exotMap
  };
  const onExotErrorGroup = onExotError || getOnExotError && getOnExotError(exotGroup) || down;
  return exotGroup;
};
/** @type { (value: unknown) => value is Exot } */ const isExot = value => isObjectAlike(value) && isString(value.id) && isBasicFunction(value.up) && isBasicFunction(value.down) && isBasicFunction(value.isUp)
/** @typedef { Map<string, Exot> } ExotMap */
/** @type { (...exotList: Exot[]) => ExotMap } */;
const toExotMap = (...exotList) => {
  /** @type { ExotMap } */
  const exotMap = new Map;
  for (const exot of exotList) exotMap.set(exot.id, exot);
  return exotMap;
};
/** @type { <T> (exotMap: ExotMap, func: (Exot) => T) => T[] } */ const mapExotMapValue = (exotMap, func) => {
  const resultList = [];
  for (const exot of exotMap.values()) resultList.push(func(exot));
  return resultList;
};
/** @type { <T> (exotMap: ExotMap, func: (Exot) => T) => T | undefined } */ const findExotMapValue = (exotMap, func) => {
  for (const exot of exotMap.values()) {
    const result = func(exot);
    if (result) return result;
  }
};
export { createExotError, createDummyExot, createExotGroup, isExot, toExotMap, mapExotMapValue, findExotMapValue };