import { tryRequire } from "../env/tryRequire.js";
import { string, basicFunction } from "./verify.js";
import { isString, isBasicFunction } from "./check.js";
import { catchSync, catchAsync } from "./error.js";
import { clock } from "./time.js";
import { time } from "./format.js";
import { indentLine } from "./string.js";
import { withTimeoutAsync } from "./function.js";
import { createTreeDepthFirstSearchAsync } from "./data/Tree.js";
const LIST_NAME_BEFORE = "beforeList";
const LIST_NAME_IT = "itList";
const LIST_NAME_AFTER = "afterList";
const createScope = (upperScope = null, level = 0, title = "") => ({
  upperScope,
  level,
  title,
  [LIST_NAME_BEFORE]: [],
  [LIST_NAME_IT]: [],
  [LIST_NAME_AFTER]: []
});
const walkScopeAsync = createTreeDepthFirstSearchAsync((scope => scope.func ? [] : [ ...scope.beforeList, ...scope.itList, ...scope.afterList ]));
const getTitleStack = (title, upperScope) => {
  const titleStack = [ title ];
  for (;upperScope; ) {
    titleStack.unshift(upperScope.title);
    upperScope = upperScope.upperScope;
  }
  return titleStack;
};
const createTest = ({colorFallback = text => text, colorError = colorFallback, colorMain = colorFallback, colorTitle = colorFallback, colorText = colorFallback, colorTime = colorFallback} = {}) => {
  let ROOT_SCOPE;
  let CURRENT_SCOPE;
  let RESULT;
  let CONFIG;
  const describe = (title, setupFunc) => {
    string(title, "invalid describe title");
    basicFunction(setupFunc, "invalid describe setupFunc");
    if (void 0 === CURRENT_SCOPE) throw new Error(`should add "describe()" between "TEST_SETUP()" & "TEST_RUN()": ${title}`);
    const innerScope = createScope(CURRENT_SCOPE, CURRENT_SCOPE.level + 1, title);
    CURRENT_SCOPE.itList.push(innerScope);
    CURRENT_SCOPE = innerScope;
    const {result, error} = catchSync(setupFunc);
    if (result instanceof Promise) throw new Error(`unexpect Promise from describe: ${title}`);
    if (error) {
      CONFIG.log(colorError(`error from describe: ${title}`));
      throw error;
    }
    CURRENT_SCOPE = CURRENT_SCOPE.upperScope;
  };
  const getPushScopeFunc = (type, typeListName) => (title, func) => {
    void 0 === func && isBasicFunction(title) && ([title, func] = [ func, title ]);
    isString(title) || (title = `${CURRENT_SCOPE.title} - ${type} #${CURRENT_SCOPE[typeListName].length}`);
    string(title, `invalid ${type} title`);
    basicFunction(func, `invalid ${type} func`);
    if (void 0 === CURRENT_SCOPE) throw new Error(`should "${type}()" between "TEST_SETUP()" & "TEST_RUN()": ${title}`);
    CURRENT_SCOPE[typeListName].push({
      upperScope: CURRENT_SCOPE,
      level: CURRENT_SCOPE.level + 1,
      title,
      func
    });
  };
  const before = getPushScopeFunc("before", "beforeList");
  const it = getPushScopeFunc("it", "itList");
  const after = getPushScopeFunc("after", "afterList");
  const stringify = (tryRequire("node:util") || {}).inspect || String;
  const info = (...args) => CONFIG && CONFIG.log(indentLine(args.map((v => isString(v) ? v : stringify(v))).join(" "), "      > "));
  return {
    TEST_SETUP: ({log = console.log, logLevel = (level, ...args) => log(indentLine(args.join(" "), "  ".repeat(level))), timeout = 42e3, testGlobal = globalThis, isSkipGlobalAssign = !1} = {}) => {
      ROOT_SCOPE = CURRENT_SCOPE = createScope(null, 0, "root");
      RESULT = {
        passList: [],
        failList: []
      };
      CONFIG = {
        log,
        logLevel,
        timeout
      };
      !isSkipGlobalAssign && Object.assign(testGlobal, {
        describe,
        it,
        before,
        after,
        info
      });
      CONFIG.log(colorMain("[TEST] setup"));
    },
    TEST_RUN: async () => {
      CONFIG.log(colorMain("[TEST] run"));
      const scope = ROOT_SCOPE;
      ROOT_SCOPE = CURRENT_SCOPE = void 0;
      const runStart = clock();
      await walkScopeAsync(scope, (async ({upperScope, title, level, func}) => {
        if (func) {
          const funcStart = clock();
          const {error} = await catchAsync(withTimeoutAsync, func, CONFIG.timeout);
          const funcDuration = clock() - funcStart;
          const timeLog = funcDuration > 64 ? colorTime(`(${time(funcDuration)})`) : "";
          if (error) {
            RESULT.failList.push({
              titleStack: getTitleStack(title, upperScope),
              error
            });
            CONFIG.logLevel(level, colorError(`[FAIL] ${title}`), timeLog);
            CONFIG.logLevel(level, colorError(indentLine(error.stack || error, "    ")));
          } else {
            RESULT.passList.push({
              title
            });
            CONFIG.logLevel(level, colorText(title), timeLog);
          }
        } else CONFIG.logLevel(level, colorTitle(title));
      }));
      CONFIG.log(colorMain([ `[TEST] done in ${time(clock() - runStart)}`, `  pass: ${RESULT.passList.length}`, `  fail: ${RESULT.failList.length}` ].join("\n")));
      for (let index = 0, indexMax = RESULT.failList.length; index < indexMax; index++) {
        const {titleStack, error} = RESULT.failList[index];
        CONFIG.log(colorError(`[FAIL|${index}]`));
        titleStack.forEach(((title, level) => level && CONFIG.logLevel(level, colorError(title))));
        CONFIG.logLevel(titleStack.length, colorError(indentLine(error.stack || error, "    ")));
      }
      return RESULT;
    },
    describe,
    it,
    before,
    after,
    info
  };
};
export { createTest };