import { Readable } from "node:stream";
import { createInterface } from "node:readline";
import { isBoolean, isObjectAlike, isBasicFunction } from "../../common/check.js";
import { createInsideOutPromise } from "../../common/function.js";
import { END, SKIP, REDO, createPack, createRunlet, createCountPool, KEY_POOL_IO, KEY_PEND_INPUT, KEY_PEND_OUTPUT, PoolIO, toPoolMap, toChipMap, toLinearChipList, quickConfigPend } from "../../common/module/Runlet.js";
const isStream = stream => isObjectAlike(stream) && isBasicFunction(stream.pipe);
const isReadableStream = stream => isStream(stream) && isBoolean(stream.readable) && isBasicFunction(stream._read);
const isStreamReadable = stream => isReadableStream(stream) && !1 !== stream.readable;
const isWritableStream = stream => isStream(stream) && isBoolean(stream.writable) && isBasicFunction(stream._write);
const isStreamWritable = stream => isWritableStream(stream) && !1 !== stream.writable;
const setupStreamPipe = (...streamList) => {
  if (streamList.length < 2) throw new Error("need at least 2 stream in streamList");
  if (!isStreamReadable(streamList[0])) throw new Error("first stream not readable");
  const lastStream = streamList[streamList.length - 1];
  const passError = error => lastStream.emit("error", error);
  for (let index = streamList.length - 2; index >= 0; index--) streamList[index].on("error", passError).pipe(streamList[index + 1]);
  return lastStream;
};
const waitStreamStopAsync = stream => new Promise(((resolve, reject) => {
  isStream(stream) || reject(new Error("expect stream"));
  stream.destroyed && reject(new Error("stream already destroyed"));
  stream.on("error", reject);
  stream.on("close", (() => reject(new Error("unexpected stream close"))));
  stream.on("end", resolve);
  stream.on("finish", resolve);
}));
const bufferToReadableStream = buffer => {
  const readableStream = new Readable;
  readableStream.push(buffer);
  readableStream.push(null);
  return readableStream;
};
const readableStreamToBufferAsync = readableStream => new Promise(((resolve, reject) => {
  isStreamReadable(readableStream) || reject(new Error("expect stream is readable"));
  const data = [];
  readableStream.on("error", reject);
  readableStream.on("close", (() => reject(new Error("unexpected stream close"))));
  readableStream.on("end", (() => resolve(Buffer.concat(data))));
  readableStream.on("data", (chunk => data.push(chunk)));
}));
const writeBufferToStreamAsync = (writableStream, buffer) => new Promise(((resolve, reject) => {
  if (0 === buffer.length) return resolve();
  writableStream.once("error", reject);
  writableStream.write(buffer, (() => {
    writableStream.off("error", reject);
    resolve();
  }));
}));
const readlineOfStreamAsync = (readableStream, onLineStringSync) => new Promise(((resolve, reject) => {
  isStreamReadable(readableStream) || reject(new Error("expect stream is readable"));
  const readlineInterface = createInterface({
    input: readableStream,
    crlfDelay: 1 / 0
  });
  const doReject = error => {
    reject(error);
    readlineInterface.close();
  };
  readableStream.on("error", doReject);
  readlineInterface.on("error", doReject);
  readlineInterface.on("close", resolve);
  readlineInterface.on("line", (lineString => {
    try {
      onLineStringSync(lineString);
    } catch (error) {
      doReject(error);
    }
  }));
}));
const createReadableStreamInputChip = ({stream, readSize, key = "chip:input:stream-readable", ...extra}) => {
  const state = {
    error: void 0,
    isEnd: !1,
    isReadable: !1,
    eventIOP: void 0
  };
  const onError = error => {
    state.error = error;
    void 0 !== state.eventIOP && state.eventIOP.reject(error);
  };
  const onEnd = () => {
    state.isEnd = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  const onReadable = () => {
    state.isReadable = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  stream.on("error", onError);
  stream.on("end", onEnd);
  stream.on("readable", onReadable);
  const allOff = () => {
    stream.off("error", onError);
    stream.off("end", onEnd);
    stream.off("readable", onReadable);
  };
  return {
    ...extra,
    key,
    prevPoolKey: KEY_POOL_IO,
    prevPendKey: KEY_PEND_INPUT,
    stream,
    state,
    process: async (pack, state, error) => {
      if (error) return allOff();
      if (void 0 !== state.error) throw state.error;
      if (!1 === state.isEnd && !1 === state.isReadable) {
        void 0 === state.eventIOP && (state.eventIOP = createInsideOutPromise());
        await state.eventIOP.promise;
        state.eventIOP = void 0;
      }
      if (!1 === state.isReadable && !0 === state.isEnd) {
        allOff();
        pack[1] = END;
      } else if (!0 === state.isReadable) {
        state.isReadable = !1;
        const chunkList = [];
        let chunk;
        for (;null !== (chunk = stream.read(readSize)); ) chunkList.push(chunk);
        pack[0] = 1 === chunkList.length ? chunkList[0] : Buffer.concat(chunkList);
        pack[1] = void 0;
      }
      return {
        pack,
        state
      };
    }
  };
};
const createWritableStreamOutputChip = ({stream, keepOpen = [ process.stdout, process.stderr ].includes(stream), IOP = createInsideOutPromise(), key = "chip:output:stream-writable", ...extra}) => {
  const state = {
    error: void 0,
    isFinish: !1,
    isDrain: !0,
    eventIOP: void 0
  };
  const onError = error => {
    state.error = error;
    void 0 !== state.eventIOP && state.eventIOP.reject(error);
  };
  const onFinish = () => {
    state.isFinish = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  const onDrain = () => {
    state.isDrain = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  stream.on("error", onError);
  stream.on("finish", onFinish);
  stream.on("drain", onDrain);
  const allOff = () => {
    stream.off("error", onError);
    stream.off("finish", onFinish);
    stream.off("drain", onDrain);
  };
  const waitIOP = async () => {
    state.eventIOP = createInsideOutPromise();
    await state.eventIOP.promise;
    state.eventIOP = void 0;
  };
  return {
    ...extra,
    key,
    nextPoolKey: KEY_POOL_IO,
    nextPendKey: KEY_PEND_OUTPUT,
    stream,
    state,
    process: async (pack, state, error) => {
      if (error) {
        allOff();
        return IOP.reject(error);
      }
      if (void 0 !== state.error) throw state.error;
      if (pack[1] === END) {
        !keepOpen && stream.end();
        for (;!1 === (keepOpen ? state.isDrain : state.isFinish); ) await waitIOP();
        allOff();
        IOP.resolve();
      } else {
        !1 === state.isDrain && await waitIOP();
        state.isDrain = stream.write(pack[0]);
        pack[0] = void 0;
        pack[1] = SKIP;
      }
      return {
        pack,
        state
      };
    },
    promise: IOP.promise
  };
};
const createTransformStreamChip = ({stream, readSize, key = "chip:stream-transform", ...extra}) => {
  const state = {
    error: void 0,
    isDrain: !0,
    isWriteEnd: !1,
    isEnd: !1,
    isReadable: !1,
    eventIOP: void 0
  };
  const onError = error => {
    state.error = error;
    void 0 !== state.eventIOP && state.eventIOP.reject(error);
  };
  const onDrain = () => {
    state.isDrain = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  const onEnd = () => {
    state.isEnd = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  const onReadable = () => {
    state.isReadable = !0;
    void 0 !== state.eventIOP && state.eventIOP.resolve();
  };
  stream.on("error", onError);
  stream.on("drain", onDrain);
  stream.on("end", onEnd);
  stream.on("readable", onReadable);
  const allOff = () => {
    stream.off("error", onError);
    stream.off("drain", onDrain);
    stream.off("end", onEnd);
    stream.off("readable", onReadable);
  };
  return {
    ...extra,
    key,
    stream,
    state,
    process: async (pack, state, error) => {
      if (error) return allOff();
      if (void 0 !== state.error) throw state.error;
      if (pack[1] === END && !1 === state.isWriteEnd) {
        state.isWriteEnd = !0;
        stream.end();
      }
      if ((!0 === state.isWriteEnd || !1 === state.isDrain) && !1 === state.isEnd && !1 === state.isReadable) {
        void 0 === state.eventIOP && (state.eventIOP = createInsideOutPromise());
        await state.eventIOP.promise;
        state.eventIOP = void 0;
      }
      let needWriteREDO = !0;
      if (pack[1] !== END && !0 === state.isDrain) {
        state.isDrain = stream.write(pack[0]);
        needWriteREDO = !1;
      }
      if (!1 === state.isReadable && !0 === state.isEnd) allOff(); else if (!0 === state.isReadable) {
        state.isReadable = !1;
        const chunkList = [];
        let chunk;
        for (;null !== (chunk = stream.read(readSize)); ) chunkList.push(chunk);
        const value = 1 === chunkList.length ? chunkList[0] : Buffer.concat(chunkList);
        if (!0 === needWriteREDO) pack = createPack(value, REDO); else {
          pack[0] = value;
          pack[1] = void 0;
        }
      } else {
        pack[0] = void 0;
        pack[1] = SKIP;
      }
      return {
        pack,
        state
      };
    }
  };
};
const quickRunletFromStream = (...streamList) => {
  if (streamList.length < 2) throw new Error("need at least 2 stream in streamList");
  const readableStream = streamList.shift();
  const writableStream = streamList.pop();
  const poolMap = toPoolMap([ PoolIO, createCountPool({
    sizeLimit: 8
  }) ]);
  const chipMap = toChipMap(toLinearChipList([ createReadableStreamInputChip({
    stream: readableStream
  }), ...streamList.map(((stream, index) => createTransformStreamChip({
    stream,
    key: `chip:transform-${index}`
  }))), createWritableStreamOutputChip({
    key: "out",
    stream: writableStream
  }) ]));
  const {attach, trigger, describe} = createRunlet(quickConfigPend(poolMap, chipMap));
  attach();
  trigger();
  return chipMap.get("out").promise;
};
export { isStream, isReadableStream, isStreamReadable, isWritableStream, isStreamWritable, setupStreamPipe, waitStreamStopAsync, bufferToReadableStream, readableStreamToBufferAsync, writeBufferToStreamAsync, readlineOfStreamAsync, createReadableStreamInputChip, createWritableStreamOutputChip, createTransformStreamChip, quickRunletFromStream };