import { join, dirname, basename } from "node:path";
import { readdirSync, mkdirSync, promises as fsAsync } from "node:fs";
import { tmpdir } from "node:os";
import { catchSync, catchAsync } from "../../common/error.js";
import { getRandomId } from "../../common/math/random.js";
import { createTreeBreadthFirstSearch, createTreeBreadthFirstSearchAsync, createTreeBottomUpSearch, createTreeBottomUpSearchAsync } from "../../common/data/Tree.js";
import { STAT_ERROR, PATH_TYPE, getPathTypeFromStat, dropTrailingSep, getPathStat, getPathStatSync, copyPath, copyPathSync, renamePath, renamePathSync, deletePath, deletePathSync } from "./Path.js";
const getPathTypeFromDirent = dirent => dirent.isSymbolicLink() ? PATH_TYPE.Symlink : dirent.isDirectory() ? PATH_TYPE.Directory : dirent.isFile() ? PATH_TYPE.File : PATH_TYPE.Other
/** @typedef { { type: PathType, name: string, path: string } } DirInfo */;
const getDirInfoList = async path => {
  const dirInfoList = [];
  for (const dirent of await fsAsync.readdir(path, {
    withFileTypes: !0
  })) {
    const type = getPathTypeFromDirent(dirent);
    const {name} = dirent;
    dirInfoList.push({
      type,
      name,
      path: join(path, name)
    });
  }
  return dirInfoList;
};
const getDirInfoListSync = path => {
  const dirInfoList = [];
  for (const dirent of readdirSync(path, {
    withFileTypes: !0
  })) {
    const type = getPathTypeFromDirent(dirent);
    const {name} = dirent;
    dirInfoList.push({
      type,
      name,
      path: join(path, name)
    });
  }
  return dirInfoList;
};
const getDirInfoTree = async path => {
  path = dropTrailingSep(path);
  const dirInfoListMap = new Map;
  const queue = [ path ];
  for (;queue.length; ) {
    const upperPath = queue.shift();
    const dirInfoList = await getDirInfoList(upperPath);
    dirInfoListMap.set(upperPath, dirInfoList);
    dirInfoList.forEach((({type, path}) => type === PATH_TYPE.Directory && queue.push(path)));
  }
  return {
    root: path,
    dirInfoListMap
  };
};
const getDirInfoTreeSync = path => {
  path = dropTrailingSep(path);
  const dirInfoListMap = new Map;
  const queue = [ path ];
  for (;queue.length; ) {
    const upperPath = queue.shift();
    const dirInfoList = getDirInfoListSync(upperPath);
    dirInfoListMap.set(upperPath, dirInfoList);
    dirInfoList.forEach((({type, path}) => type === PATH_TYPE.Directory && queue.push(path)));
  }
  return {
    root: path,
    dirInfoListMap
  };
};
const getSubNodeListFunc = (dirInfo, dirInfoListMap) => dirInfoListMap.get(dirInfo.path);
const dirInfoTreeBreadthFirstSearchAsync = createTreeBreadthFirstSearchAsync(getSubNodeListFunc);
const dirInfoTreeBreadthFirstSearchSync = createTreeBreadthFirstSearch(getSubNodeListFunc);
const dirInfoTreeBottomUpSearchAsync = createTreeBottomUpSearchAsync(getSubNodeListFunc);
const dirInfoTreeBottomUpSearchSync = createTreeBottomUpSearch(getSubNodeListFunc);
const walkDirInfoTree = async ({root, dirInfoListMap}, callback) => dirInfoTreeBreadthFirstSearchAsync({
  path: root
}, callback, dirInfoListMap);
const walkDirInfoTreeSync = ({root, dirInfoListMap}, callback) => dirInfoTreeBreadthFirstSearchSync({
  path: root
}, callback, dirInfoListMap);
const walkDirInfoTreeBottomUp = async ({root, dirInfoListMap}, callback) => dirInfoTreeBottomUpSearchAsync({
  path: root
}, callback, dirInfoListMap);
const walkDirInfoTreeBottomUpSync = ({root, dirInfoListMap}, callback) => dirInfoTreeBottomUpSearchSync({
  path: root
}, callback, dirInfoListMap);
const copyDirInfoTree = async (dirInfoTree, pathTo) => {
  await createDirectory(pathTo);
  const pathToMap = new Map;
  pathToMap.set(dirInfoTree.root, pathTo);
  return walkDirInfoTree(dirInfoTree, (async ({name, path}) => {
    const upperPath = dirname(path);
    const pathTo = join(pathToMap.get(upperPath), name);
    pathToMap.set(path, pathTo);
    await copyPath(path, pathTo);
  }));
};
const copyDirInfoTreeSync = (dirInfoTree, pathTo) => {
  createDirectorySync(pathTo);
  const pathToMap = new Map;
  pathToMap.set(dirInfoTree.root, pathTo);
  return walkDirInfoTreeSync(dirInfoTree, (({name, path}) => {
    const upperPath = dirname(path);
    const pathTo = join(pathToMap.get(upperPath), name);
    pathToMap.set(path, pathTo);
    copyPathSync(path, pathTo);
  }));
};
const renameDirInfoTree = async ({root, dirInfoListMap}, pathTo) => {
  await createDirectory(pathTo);
  for (const {name, path} of dirInfoListMap.get(root)) await renamePath(path, join(pathTo, name));
};
const renameDirInfoTreeSync = ({root, dirInfoListMap}, pathTo) => {
  createDirectorySync(pathTo);
  for (const {name, path} of dirInfoListMap.get(root)) renamePathSync(path, join(pathTo, name));
};
const deleteDirInfoTree = async dirInfoListMap => walkDirInfoTreeBottomUp(dirInfoListMap, (({path}) => deletePath(path)));
const deleteDirInfoTreeSync = dirInfoListMap => walkDirInfoTreeBottomUpSync(dirInfoListMap, (({path}) => deletePathSync(path)));
const createDirectory = async (path, pathStat) => {
  if (void 0 !== pathStat) {
    if (pathStat.isDirectory()) return;
    if (pathStat !== STAT_ERROR) throw new Error(`path already taken by non-directory: ${path}`);
  }
  await fsAsync.mkdir(path, {
    recursive: !0
  });
};
const createDirectorySync = (path, pathStat) => {
  if (void 0 !== pathStat) {
    if (pathStat.isDirectory()) return;
    if (pathStat !== STAT_ERROR) throw new Error(`path already taken by non-directory: ${path}`);
  }
  mkdirSync(path, {
    recursive: !0
  });
};
const copyDirectory = async (pathFrom, pathTo, pathStat) => copyDirInfoTree(await getDirInfoTree(pathFrom), pathTo);
const copyDirectorySync = (pathFrom, pathTo, pathStat) => copyDirInfoTreeSync(getDirInfoTreeSync(pathFrom), pathTo);
const deleteDirectory = async (path, pathStat) => {
  await deleteDirInfoTree(await getDirInfoTree(path));
  await deletePath(path, pathStat);
};
const deleteDirectorySync = (path, pathStat) => {
  deleteDirInfoTreeSync(getDirInfoTreeSync(path));
  deletePathSync(path, pathStat);
};
const resetDirectory = async (path, pathStat) => {
  void 0 === pathStat && (pathStat = await getPathStat(path));
  if (pathStat.isDirectory()) await deleteDirInfoTree(await getDirInfoTree(path)); else {
    pathStat !== STAT_ERROR && await deletePath(path, pathStat);
    await createDirectory(path);
  }
};
const resetDirectorySync = (path, pathStat) => {
  void 0 === pathStat && (pathStat = getPathStatSync(path));
  if (pathStat.isDirectory()) deleteDirInfoTreeSync(getDirInfoTreeSync(path)); else {
    pathStat !== STAT_ERROR && deletePathSync(path, pathStat);
    createDirectorySync(path);
  }
};
const withTempDirectory = async (asyncFunc, pathTemp = join(tmpdir(), "dr-js", getRandomId("temp-"))) => {
  await resetDirectory(pathTemp);
  const {result, error} = await catchAsync(asyncFunc, pathTemp);
  await deleteDirectory(pathTemp);
  if (error) throw error;
  return result;
};
/** @type { <T> (syncFunc: (pathTemp: string) => T, pathTemp?: string) => T } */ const withTempDirectorySync = (syncFunc, pathTemp = join(tmpdir(), "dr-js", getRandomId("temp-"))) => {
  resetDirectorySync(pathTemp);
  const {result, error} = catchSync(syncFunc, pathTemp);
  deleteDirectorySync(pathTemp);
  if (error) throw error;
  return result;
};
/** @typedef { (fileList: string[], dirInfo: DirInfo) => Promise<void> } FileCollectorAsync */
/** @type { (path: string, fileCollector: FileCollectorAsync) => Promise<string[]> } */ const getFileList = async (path, fileCollector = (async (fileList, {path}) => {
  fileList.push(path);
})) => {
  const fileList = [];
  const pathStat = await getPathStat(path);
  const pathType = getPathTypeFromStat(pathStat);
  switch (pathType) {
   case PATH_TYPE.File:
    await fileCollector(fileList, {
      type: pathType,
      name: basename(path),
      path
    });
    break;
   case PATH_TYPE.Directory:
    await walkDirInfoTree(await getDirInfoTree(path), (dirInfo => dirInfo.type === PATH_TYPE.File && fileCollector(fileList, dirInfo)));
    break;
   default:
    throw new Error(`invalid pathType: ${pathType} for ${path}`);
  }
  return fileList;
};
/** @typedef { (fileList: string[], dirInfo: DirInfo) => void } FileCollector */
/** @type { (path: string, fileCollector: FileCollector) => string[] } */ const getFileListSync = (path, fileCollector = ((fileList, {path}) => {
  fileList.push(path);
})) => {
  const fileList = [];
  const pathStat = getPathStatSync(path);
  const pathType = getPathTypeFromStat(pathStat);
  switch (pathType) {
   case PATH_TYPE.File:
    fileCollector(fileList, {
      type: pathType,
      name: basename(path),
      path
    });
    break;
   case PATH_TYPE.Directory:
    walkDirInfoTreeSync(getDirInfoTreeSync(path), (dirInfo => dirInfo.type === PATH_TYPE.File && fileCollector(fileList, dirInfo)));
    break;
   default:
    throw new Error(`invalid pathType: ${pathType} for ${path}`);
  }
  return fileList;
};
export { getPathTypeFromDirent, getDirInfoList, getDirInfoListSync, getDirInfoTree, getDirInfoTreeSync, walkDirInfoTree, walkDirInfoTreeSync, walkDirInfoTreeBottomUp, walkDirInfoTreeBottomUpSync, copyDirInfoTree, copyDirInfoTreeSync, renameDirInfoTree, renameDirInfoTreeSync, deleteDirInfoTree, deleteDirInfoTreeSync, createDirectory, createDirectorySync, copyDirectory, copyDirectorySync, deleteDirectory, deleteDirectorySync, resetDirectory, resetDirectorySync, withTempDirectory, withTempDirectorySync, getFileList, getFileListSync };