import { resolve } from "node:path";
import { homedir, tmpdir } from "node:os";
import { clock } from "../common/time.js";
import { time } from "../common/format.js";
import { clamp } from "../common/math/base.js";
import { findUpPackageRoot, getSudoArgs } from "./module/Software/npm.js";
import { createColor } from "./module/TerminalTTY.js";
import { resolveCommand } from "./system/ResolveCommand.js";
import { runDetached, runSync } from "./run.js";
const loadEnvKey = key => {
  try {
    return JSON.parse(process.env[key]);
  } catch (error) {
    return null;
  }
};
const saveEnvKey = (key, value) => {
  try {
    process.env[key] = JSON.stringify(value);
  } catch (error) {}
};
const syncEnvKey = (key, defaultValue) => {
  const value = loadEnvKey(key) || defaultValue;
  saveEnvKey(key, value);
  return value;
};
const argvFlag = (...checkFlagList) => process.argv.find((flag => checkFlagList.includes(flag)));
const ENV_KEY_LOGGER = "__DRJS_KIT_LOGGER__";
const ENV_KEY_VERBOSE = "__DRJS_KIT_VERBOSE__";
const getKitLogger = ({title = "kit", isNoEnvKey = !1, isVerbose = Boolean(loadEnvKey("__DRJS_KIT_VERBOSE__") || argvFlag("verbose") || process.env.KIT_VERBOSE), isQuiet = Boolean(argvFlag("quiet") || process.env.KIT_QUIET), padWidth = clamp(process.env.KIT_LOGGER_WIDTH || process.stdout.isTTY && process.stdout.columns || 80, 32, 240), TerminalColor = createColor(), colorPadLogFunc = TerminalColor.fg.yellow, colorStepLogFunc = TerminalColor.fg.yellow, colorLogFunc = TerminalColor.fg.darkGray, logFunc = console.log} = {}) => {
  const startTime = clock();
  !isNoEnvKey && saveEnvKey("__DRJS_KIT_VERBOSE__", Boolean(isVerbose));
  const envKit = loadEnvKey("__DRJS_KIT_LOGGER__") || {
    titleList: [],
    startTime,
    pid: null
  };
  !isNoEnvKey && envKit.pid !== process.pid && saveEnvKey("__DRJS_KIT_LOGGER__", {
    ...envKit,
    titleList: [ title, ...envKit.titleList ],
    pid: process.pid
  });
  title = [ title, ...envKit.titleList.map((v => v.slice(0, clamp(Math.floor(padWidth / 20), 3, 9)))) ].join("|");
  let prevTime = clock();
  const stepLog = (...args) => logFunc(`- ${colorStepLogFunc(`(+${(() => {
    const currentTime = clock();
    const stepTime = currentTime - prevTime;
    prevTime = currentTime;
    return time(stepTime);
  })()})`)} ${args.join(" ")}`);
  const log = (...args) => logFunc(colorLogFunc(`- ${args.join(" ")}`));
  const devLog = isVerbose ? log : () => {};
  return !isVerbose && isQuiet ? {
    padLog: stepLog,
    stepLog: devLog,
    log: devLog,
    devLog: () => {},
    isVerbose
  } : {
    padLog: (...args) => {
      const start = `## ${args.join(" ")} `;
      const end = ` [${title}|${(() => {
        const currentTime = clock();
        prevTime = currentTime;
        return envKit.startTime === startTime ? time(currentTime - startTime) : `${time(currentTime - startTime)}/${time(currentTime - envKit.startTime)}`;
      })()}]`;
      logFunc(`\n${start.padEnd(padWidth - end.length, "-")}${colorPadLogFunc(end)}`);
    },
    stepLog,
    log,
    devLog,
    isVerbose
  };
};
const getKitPathCombo = ({PATH_ROOT = findUpPackageRoot(process.cwd()) || process.cwd(), PATH_OUTPUT = "output-gitignore/", PATH_TEMP = ".temp-gitignore/", PATH_HOME = homedir(), PATH_OSTEMP = tmpdir()} = {}) => {
  PATH_OUTPUT = resolve(PATH_ROOT, PATH_OUTPUT);
  PATH_TEMP = resolve(PATH_ROOT, PATH_TEMP);
  PATH_HOME = resolve(PATH_ROOT, PATH_HOME);
  PATH_OSTEMP = resolve(PATH_ROOT, PATH_OSTEMP);
  return {
    PATH_ROOT,
    fromRoot: (...args) => resolve(PATH_ROOT, ...args),
    PATH_OUTPUT,
    fromOutput: (...args) => resolve(PATH_OUTPUT, ...args),
    PATH_TEMP,
    fromTemp: (...args) => resolve(PATH_TEMP, ...args),
    PATH_HOME,
    fromHome: (...args) => resolve(PATH_HOME, ...args),
    PATH_OSTEMP,
    fromOsTemp: (...args) => resolve(PATH_OSTEMP, ...args)
  };
};
const getKitRun = ({PATH_ROOT, log = console.log, isQuiet = Boolean(argvFlag("quiet") || process.env.KIT_QUIET), isDryRun = Boolean(process.env.KIT_DRY_RUN)} = {}) => {
  const toArgList = argListOrString => Array.isArray(argListOrString) ? [ ...argListOrString ] : argListOrString.split(" ").filter(Boolean);
  const RUN = (argListOrString, {isDetached = !1, ...option} = {}) => {
    const argList = toArgList(argListOrString);
    argList[0] = resolveCommand(argList[0], PATH_ROOT);
    log(`[${isDryRun ? "DRY-RUN" : "RUN"}${isDetached ? "|DETACHED" : ""}] "${argList.join(" ")}"`);
    if (!isDryRun) return (isDetached ? runDetached : runSync)(argList, {
      cwd: PATH_ROOT,
      quiet: isQuiet,
      ...option
    });
  };
  return {
    RUN,
    RUN_SUDO_NPM: (argListOrString, option) => RUN([ ...getSudoArgs(), ...toArgList(argListOrString) ], option)
  };
};
const getKit = option => {
  const kitLogger = getKitLogger(option);
  const kitPathCombo = getKitPathCombo(option);
  const kitRun = getKitRun({
    ...option,
    ...kitLogger,
    ...kitPathCombo
  });
  return {
    ...kitLogger,
    ...kitPathCombo,
    ...kitRun
  };
};
const runKit = (asyncFunc, {kit, kitLogger, title = process.argv.slice(2).join("+") || void 0, ...kitOption} = {}) => {
  const kitOrKitLogger = kit || kitLogger || getKit({
    title,
    ...kitOption
  });
  const startTime = clock();
  new Promise((resolve => resolve(asyncFunc(kitOrKitLogger)))).then((() => {
    kitOrKitLogger.padLog(`done in ${time(clock() - startTime)}`);
  }), (error => {
    console.warn(error);
    kitOrKitLogger.padLog(`error after ${time(clock() - startTime)}`);
    process.exit(-1);
  }));
};
export { loadEnvKey, saveEnvKey, syncEnvKey, argvFlag, ENV_KEY_LOGGER, ENV_KEY_VERBOSE, getKitLogger, getKitPathCombo, getKitRun, getKit, runKit };