import { posix, relative } from "node:path";
import { promises as fsAsync } from "node:fs";
import { catchAsync } from "../../../common/error.js";
import { objectMap } from "../../../common/immutable/Object.js";
import { PATH_TYPE, getPathStat, existPath, toPosixPath, createPathPrefixLock } from "../../fs/Path.js";
import { getDirInfoList, getDirInfoTree, walkDirInfoTree, createDirectory } from "../../fs/Directory.js";
import { modifyRename, modifyCopy, modifyDelete } from "../../fs/Modify.js";
const PATH_VISIBLE = "path.visible";
const PATH_STAT = "path.stat";
const PATH_COPY = "path.copy";
const PATH_RENAME = "path.rename";
const PATH_DELETE = "path.delete";
const PATH_DIRECTORY_CREATE = "path.dir.create";
const PATH_DIRECTORY_CONTENT = "path.dir.content";
const PATH_DIRECTORY_ALL_FILE_LIST = "path.dir.all-file-list";
const ACTION_TYPE = {
  PATH_VISIBLE: "path.visible",
  PATH_STAT: "path.stat",
  PATH_COPY: "path.copy",
  PATH_RENAME: "path.rename",
  PATH_DELETE: "path.delete",
  PATH_DIRECTORY_CREATE: "path.dir.create",
  PATH_DIRECTORY_CONTENT: "path.dir.content",
  PATH_DIRECTORY_ALL_FILE_LIST: "path.dir.all-file-list"
};
const ACTION_CORE_MAP = {
  [PATH_VISIBLE]: async absolutePath => existPath(absolutePath).then((isVisible => ({
    isVisible
  }))),
  [PATH_STAT]: async absolutePath => fsAsync.stat(absolutePath).then((({mode, size, mtimeMs}) => ({
    mode,
    size,
    mtimeMs
  }))),
  [PATH_COPY]: async (absolutePath, absolutePathTo) => {
    await modifyCopy(absolutePath, absolutePathTo);
  },
  [PATH_RENAME]: async (absolutePath, absolutePathTo) => {
    await modifyRename(absolutePath, absolutePathTo);
  },
  [PATH_DELETE]: async absolutePath => {
    await modifyDelete(absolutePath);
  },
  [PATH_DIRECTORY_CREATE]: async absolutePath => {
    await createDirectory(absolutePath);
  },
  [PATH_DIRECTORY_CONTENT]: async absolutePath => {
    const {result: dirInfoList = [], error} = await catchAsync(getDirInfoList, absolutePath);
    const directoryList = [];
    const fileList = [];
    for (const {type, name, path} of dirInfoList) {
      const stat = type === PATH_TYPE.Directory || await getPathStat(path);
      !0 === stat || stat.isDirectory() ? directoryList.push(name) : fileList.push([ name, stat.size, Math.round(stat.mtimeMs) ]);
    }
    return {
      directoryList,
      fileList
    };
  },
  [PATH_DIRECTORY_ALL_FILE_LIST]: async absolutePath => {
    const fileList = [];
    const {error} = await catchAsync((async () => walkDirInfoTree(await getDirInfoTree(absolutePath), (async ({type, path}) => {
      const stat = type === PATH_TYPE.Directory || await getPathStat(path);
      !0 !== stat && !stat.isDirectory() && fileList.push([ toPosixPath(relative(absolutePath, path)), stat.size, Math.round(stat.mtimeMs) ]);
    }))));
    return {
      fileList
    };
  }
};
const setupActionMap = ({actionCoreMap = ACTION_CORE_MAP, rootPath, getPath = createPathPrefixLock(rootPath), loggerExot}) => {
  const calcPath = (key, extraPath) => getPath(posix.normalize(posix.join(key, extraPath)));
  return objectMap(actionCoreMap, ((actionFunc, actionType) => async (store, {key, keyTo, batchList = [ "" ]}) => {
    loggerExot.add(`[ActionBox|${actionType}] ${key || ""}${keyTo ? `, ${keyTo}` : ""} (${batchList.join("|")})`);
    const resultList = [];
    const errorList = [];
    for (const extraPath of batchList) {
      const absolutePath = calcPath(key, extraPath);
      const absolutePathTo = keyTo && calcPath(keyTo, extraPath);
      const {result, error} = await catchAsync(actionFunc, absolutePath, absolutePathTo);
      const response = {
        actionType,
        key,
        keyTo,
        ...result
      };
      error ? errorList.push({
        ...response,
        error: String(error)
      }) : resultList.push(response);
    }
    return {
      resultList,
      errorList
    };
  }));
};
export { ACTION_TYPE, ACTION_CORE_MAP, setupActionMap };