import { verifyCheckCode, verifyParsedCheckCode, generateCheckCode, generateLookupData, parseCheckCode, packDataArrayBuffer, parseDataArrayBuffer } from "../../common/module/TimedLookup.js";
import { createCacheMap } from "../../common/data/CacheMap.js";
import { time, binary, prettyStringifyJSON } from "../../common/format.js";
import { indentList } from "../../common/string.js";
import { getEntityTagByContentHash } from "./EntityTag.js";
import { fetchLikeRequest } from "../net.js";
import { readArrayBuffer, writeArrayBuffer } from "../fs/File.js";
import { createPathPrefixLock } from "../fs/Path.js";
const saveAuthFile = async (pathFile, timedLookupData) => writeArrayBuffer(pathFile, packDataArrayBuffer(timedLookupData));
const loadAuthFile = async pathFile => parseDataArrayBuffer(await readArrayBuffer(pathFile));
const describeAuthFile = async pathFile => {
  const {tag, size, tokenSize, timeGap, info, dataView} = await loadAuthFile(pathFile);
  return indentList("[AuthFile]", [ `tag: ${tag}`, `size: ${binary(size)}B`, `tokenSize: ${tokenSize}B`, `timeGap: ${time(1e3 * timeGap)}`, info && `info: ${prettyStringifyJSON(info)}`, `dataViewEntityTag: ${getEntityTagByContentHash(Buffer.from(dataView.buffer))}` ].filter(Boolean));
};
const generateAuthFile = async (pathFile, {tag, size, tokenSize, timeGap, info}) => {
  const timedLookupData = await generateLookupData({
    tag,
    size,
    tokenSize,
    timeGap,
    info
  });
  await saveAuthFile(pathFile, timedLookupData);
  return timedLookupData;
};
const generateAuthCheckCode = async (pathFile, timestamp) => generateCheckCode(await loadAuthFile(pathFile), timestamp);
const verifyAuthCheckCode = async (pathFile, checkCode, timestamp) => verifyCheckCode(await loadAuthFile(pathFile), checkCode, timestamp);
const DEFAULT_AUTH_KEY = "auth-check-code";
const AUTH_SKIP = "auth-skip";
const AUTH_FILE = "auth-file";
const AUTH_FILE_GROUP = "auth-file-group";
const authFetchWrap = async (url, option = {}, timedLookupData, authKey) => {
  const response = await fetchLikeRequest(url, {
    ...option,
    headers: {
      [authKey]: generateCheckCode(timedLookupData),
      ...option.headers
    }
  });
  if (!response.ok) throw new Error(`[authFetch] status: ${response.status}`);
  return response;
};
const configureAuthSkip = async ({authKey = "auth-check-code", log}) => {
  log && log("auth skip enabled");
  return {
    authMode: "auth-skip",
    authKey,
    authFetch: fetchLikeRequest,
    checkAuth: async () => "auth-skip",
    generateAuthCheckCode: async () => "auth-skip"
  };
};
const configureAuthFile = async ({authKey = "auth-check-code", log, timedLookupData, authFile}) => {
  if (!timedLookupData) {
    timedLookupData = await loadAuthFile(authFile).catch((async error => {
      log && log("missing auth file", error);
      throw error;
    }));
    log && log("loaded auth file");
  }
  return {
    authMode: "auth-file",
    authKey,
    authFetch: async (url, option) => authFetchWrap(url, option, timedLookupData, authKey),
    checkAuth: async checkCode => {
      verifyCheckCode(timedLookupData, checkCode);
      return timedLookupData.tag;
    },
    generateAuthCheckCode: async () => generateCheckCode(timedLookupData)
  };
};
const configureAuthFileGroup = async ({authKey = "auth-check-code", log, authFileGroupPath, authFileGroupDefaultTag, authFileGroupKeySuffix, getFileNameForTag = authFileGroupKeySuffix ? tag => `${tag}${authFileGroupKeySuffix}` : tag => `${tag}.key`, authCacheExpireTime = 3e5, authCacheMap = createCacheMap({
  valueSizeSumMax: 8388608,
  expireAfter: authCacheExpireTime,
  eventHub: null
})}) => {
  const getPath = createPathPrefixLock(authFileGroupPath);
  const getTimedLookupData = tag => authCacheMap.get(tag) || loadAuthFile(getPath(getFileNameForTag(tag))).then((timedLookupData => {
    authCacheMap.set(tag, timedLookupData, timedLookupData.dataView.byteLength);
    log && log(`loaded auth file for tag: ${tag}`);
    return timedLookupData;
  }), (error => {
    log && log(`no auth file for tag: ${tag}`);
    throw error;
  }));
  return {
    authMode: "auth-file-group",
    authKey,
    authFetch: async (url, option, tag = authFileGroupDefaultTag) => authFetchWrap(url, option, await getTimedLookupData(tag), authKey),
    checkAuth: async checkCode => {
      const parsedCheckCode = parseCheckCode(checkCode);
      const tag = parsedCheckCode[0];
      const timedLookupData = await getTimedLookupData(tag);
      verifyParsedCheckCode(timedLookupData, parsedCheckCode);
      return timedLookupData.tag;
    },
    generateAuthCheckCode: async (tag = authFileGroupDefaultTag) => generateCheckCode(await getTimedLookupData(tag))
  };
};
const configureAuth = async ({authKey = "auth-check-code", log, authSkip = !1, authFile, authFileGroupPath, authFileGroupDefaultTag, authFileGroupKeySuffix}) => {
  if (authSkip) return configureAuthSkip({
    authKey,
    log
  });
  if (authFile) return configureAuthFile({
    authFile,
    authKey,
    log
  });
  if (authFileGroupPath) return configureAuthFileGroup({
    authFileGroupPath,
    authFileGroupDefaultTag,
    authFileGroupKeySuffix,
    authKey,
    log
  });
  throw new Error("no auth mode option provided");
};
export { saveAuthFile, loadAuthFile, describeAuthFile, generateAuthFile, generateAuthCheckCode, verifyAuthCheckCode, DEFAULT_AUTH_KEY, AUTH_SKIP, AUTH_FILE, AUTH_FILE_GROUP, configureAuthSkip, configureAuthFile, configureAuthFileGroup, configureAuth };