import { resolve, dirname } from "node:path";
import { createReadStream, createWriteStream, promises as fsAsync } from "node:fs";
import { rethrowError } from "../../common/error.js";
import { getRandomId } from "../../common/math/random.js";
import { createCacheMap } from "../../common/data/CacheMap.js";
import { fromNodejsBuffer } from "../../common/data/ArrayBuffer.js";
import { createAsyncFuncQueue } from "../../common/module/AsyncFuncQueue.js";
import { parseArrayBufferChunk, uploadArrayBufferByChunk } from "../../common/module/ChunkUpload.js";
import { quickRunletFromStream } from "../data/Stream.js";
import { writeBuffer } from "../fs/File.js";
import { createPathPrefixLock } from "../fs/Path.js";
import { createDirectory } from "../fs/Directory.js";
import { modifyDelete, modifyDeleteForce } from "../fs/Modify.js";
const createOnFileChunkUpload = async ({rootPath, mergePath, onError = rethrowError, allowSkipHashVerify = !0, expireTime = 6e5, chunkCacheMap = createCacheMap({
  valueSizeSumMax: 64,
  expireAfter: expireTime
})}) => {
  await createDirectory(rootPath);
  await createDirectory(mergePath);
  const getPath = createPathPrefixLock(rootPath);
  const {push} = createAsyncFuncQueue(onError);
  chunkCacheMap.subscribe((({type, key, payload}) => {
    if ("delete" !== type) return;
    const {tempPath} = payload;
    push((() => modifyDeleteForce(tempPath)));
  }));
  return async ({bufferPacket, cacheKeyPrefix = "", onUploadStart, onUploadChunk, onUploadEnd}) => {
    const {chunkArrayBuffer, key, chunkIndex, chunkTotal} = await parseArrayBufferChunk(fromNodejsBuffer(bufferPacket), allowSkipHashVerify);
    const cacheKey = `${cacheKeyPrefix}-${key}-${chunkTotal}`;
    let chunkData = chunkCacheMap.get(cacheKey);
    if (void 0 === chunkData) {
      const filePath = getPath(key);
      const tempPath = resolve(mergePath, getRandomId(cacheKey).replace(/[^\w-.]/g, "_"));
      chunkData = {
        tempPath,
        filePath,
        key,
        chunkTotal,
        chunkList: []
      };
      onUploadStart && await onUploadStart(chunkData);
      await createDirectory(tempPath);
    }
    const chunkPath = resolve(chunkData.tempPath, `chunk-${chunkIndex}-${chunkTotal}`);
    await writeBuffer(chunkPath, Buffer.from(chunkArrayBuffer));
    chunkData.chunkList[chunkIndex] = {
      chunkIndex,
      chunkPath
    };
    onUploadChunk && await onUploadChunk(chunkData, chunkIndex);
    const chunkCacheCount = Object.keys(chunkData.chunkList).length;
    if (chunkCacheCount === chunkTotal) {
      await createDirectory(dirname(chunkData.filePath));
      await writeBuffer(chunkData.filePath, "");
      for (const {chunkPath} of chunkData.chunkList) await quickRunletFromStream(createReadStream(chunkPath), createWriteStream(chunkData.filePath, {
        flags: "a"
      }));
      await modifyDelete(chunkData.tempPath);
      chunkCacheMap.delete(cacheKey);
      onUploadEnd && await onUploadEnd(chunkData);
    } else chunkCacheCount > 1 ? chunkCacheMap.touch(cacheKey) : chunkCacheMap.set(cacheKey, chunkData);
  };
};
const uploadFileByChunk = async ({fileBuffer, filePath, fileSize, key, chunkSizeMax, uploadChunk, onProgress}) => {
  void 0 === fileSize && (fileSize = fileBuffer ? fileBuffer.length : (await fsAsync.stat(filePath)).size);
  const fileHandle = fileBuffer ? void 0 : await fsAsync.open(filePath);
  const getChunk = fileBuffer ? (index, chunkSize) => fromNodejsBuffer(fileBuffer.slice(index, index + chunkSize)) : async (index, chunkSize) => {
    const buffer = Buffer.allocUnsafe(chunkSize);
    await fileHandle.read(buffer, 0, chunkSize, index);
    return fromNodejsBuffer(buffer);
  };
  await uploadArrayBufferByChunk({
    size: fileSize,
    getChunk,
    key,
    chunkSizeMax,
    uploadChunk,
    onProgress
  });
  fileHandle && await fileHandle.close();
};
export { createOnFileChunkUpload, uploadFileByChunk };