import { resolve } from "node:path";
import { createDirectory } from "../fs/Directory.js";
import { createSafeWriteStream } from "./SafeWrite.js";
/** @typedef { Exot & { add: (logString: string) => void, save: () => void } } SimpleLoggerExot */
/** @typedef { { id?: string, queueLengthThreshold?: number } & OptCreateSafeWriteStream } OptCreateSimpleLoggerExot */
/** @type { (opt: OptCreateSimpleLoggerExot) => SimpleLoggerExot } */ const createSimpleLoggerExot = ({id = "exot:logger-simple", queueLengthThreshold = 1024, ...safeWriteStreamOption}) => {
  const logQueue = [];
  let safeWriteStream;
  const isUp = () => void 0 !== safeWriteStream;
  const save = () => {
    if (!1 !== isUp() && 0 !== logQueue.length) {
      logQueue.push("");
      safeWriteStream.write(logQueue.join("\n"));
      logQueue.length = 0;
    }
  };
  return {
    id,
    up: onExotError => {
      if (!0 === isUp()) throw new Error("already up");
      safeWriteStream = createSafeWriteStream({
        onError: onExotError,
        ...safeWriteStreamOption
      });
    },
    down: () => {
      if (!1 !== isUp()) {
        save();
        safeWriteStream.end();
        safeWriteStream = void 0;
      }
    },
    isUp,
    add: logString => {
      logQueue.push(logString);
      logQueue.length > queueLengthThreshold && save();
    },
    save
  };
};
/** @typedef { SimpleLoggerExot & { add: (...args: string[]) => void, split: () => void } } LoggerExot */
/** @typedef { { id?: string, pathLogDirectory: string, getLogFileName?: () => string, saveInterval?: number, splitInterval?: number } & OptCreateSimpleLoggerExot } OptCreateLoggerExot */
/** @type { (opt: OptCreateLoggerExot) => LoggerExot } */ const createLoggerExot = ({id = "exot:logger", pathLogDirectory, getLogFileName = () => `${(new Date).toISOString().replace(/\W/g, "-")}.log`, saveInterval = 3e4, splitInterval = 864e5, ...simpleLoggerOption}) => {
  const getLoggerExot = () => createSimpleLoggerExot({
    ...simpleLoggerOption,
    pathOutputFile: resolve(pathLogDirectory, getLogFileName())
  });
  let loggerExot = getLoggerExot();
  let saveToken;
  let splitToken;
  const upSync = onExotError => {
    loggerExot.up(onExotError);
    saveToken = saveInterval ? setInterval(save, saveInterval) : void 0;
    splitToken = splitInterval ? setTimeout(split, splitInterval) : void 0;
  };
  const down = () => {
    if (!1 !== isUp()) {
      loggerExot.down();
      void 0 !== saveToken && clearInterval(saveToken);
      void 0 !== splitToken && clearTimeout(splitToken);
      loggerExot = getLoggerExot();
      saveToken = void 0;
      splitToken = void 0;
    }
  };
  const isUp = () => loggerExot.isUp();
  const save = () => {
    loggerExot.save();
  };
  const split = () => {
    if (!1 !== isUp()) {
      down();
      upSync();
    }
  };
  return {
    id,
    up: async onExotError => {
      if (!0 === isUp()) throw new Error("already up");
      await createDirectory(pathLogDirectory);
      upSync(onExotError);
    },
    down,
    isUp,
    add: (...args) => {
      loggerExot.add(args.join(" "));
    },
    save,
    split
  };
};
export { createSimpleLoggerExot, createLoggerExot };