import { request as httpRequest } from "node:http";
import { request as httpsRequest } from "node:https";
import { createGunzip } from "node:zlib";
import { clock } from "../common/time.js";
import { remessageError } from "../common/error.js";
import { isString, isArrayBuffer } from "../common/check.js";
import { createInsideOutPromise, withRetryAsync } from "../common/function.js";
import { fromNodejsBuffer } from "../common/data/ArrayBuffer.js";
import { isStream, setupStreamPipe, readableStreamToBufferAsync } from "./data/Stream.js";
const requestHttp = (url, option, body) => {
  url = url instanceof URL ? url : new URL(url);
  const method = option.method || "GET";
  const tagError = error => remessageError(Object.assign(error, {
    url,
    method
  }), `[${method}|${url}] ${error.message}`);
  const {promise, resolve, reject} = createInsideOutPromise();
  const onError = error => {
    request.destroy();
    reject(tagError(error));
  };
  const onTimeout = option.timeout && (() => onError(new Error("NETWORK_TIMEOUT")));
  const request = ("https:" === url.protocol ? httpsRequest : httpRequest)(url, option, (response => {
    request.off("error", onError);
    onTimeout && request.off("timeout", onTimeout);
    request.on("error", (error => {
      request.destroy();
      response.emit("error", tagError(error));
    }));
    resolve(response);
  }));
  request.on("error", onError);
  onTimeout && request.on("timeout", onTimeout);
  isStream(body) ? setupStreamPipe(body, request) : isArrayBuffer(body) ? request.end(Buffer.from(body)) : request.end(body);
  return {
    url,
    request,
    promise,
    tagError
  };
};
const ping = async (url, {method = "GET", headers, body, wait = 5e3, maxRetry = 0} = {}) => withRetryAsync((async () => {
  (await requestHttp(url, {
    method,
    headers,
    timeout: wait
  }, body).promise).destroy();
}), maxRetry, wait);
const fetchLikeRequest = async (url, {method = "GET", headers: requestHeaders, body, bodyLength, timeout = 1e4, onProgressUpload, onProgressDownload, ...extra} = {}) => {
  const option = {
    method,
    headers: {
      "accept-encoding": "gzip",
      ...requestHeaders
    },
    timeout,
    ...extra
  };
  const {request, promise, tagError} = requestHttp(url, option, body);
  const timeStart = clock();
  body && onProgressUpload && request.once("socket", (socket => {
    bodyLength = bodyLength || (isStream(body) ? 1 / 0 : isArrayBuffer(body) ? body.byteLength : isString(body) ? Buffer.byteLength(body) : body.length);
    const bytesWrittenStart = socket.bytesWritten;
    const onProgress = () => {
      onProgressUpload(Math.min(socket.bytesWritten - bytesWrittenStart, bodyLength), bodyLength);
    };
    socket.on("drain", onProgress);
    const clearListener = () => socket.off("drain", onProgress);
    promise.then(clearListener, clearListener);
  }));
  const response = await promise;
  const status = response.statusCode;
  return {
    status,
    ok: status >= 200 && status < 300,
    headers: response.headers,
    ..._wrapPayload(url, method, request, response, timeout + timeStart - clock(), onProgressDownload, tagError)
  };
};
const _wrapPayload = (url, method, request, response, timeoutPayload, onProgressDownload, tagError) => {
  let payloadOutcome;
  process.nextTick((() => {
    if (!payloadOutcome) {
      payloadOutcome = "DROP";
      request.destroy();
    }
  }));
  const _stream = () => {
    if (payloadOutcome) throw new Error("KEEP" === payloadOutcome ? "PAYLOAD_ALREADY_USED" : "PAYLOAD_ALREADY_DROPPED");
    payloadOutcome = "KEEP";
    if (timeoutPayload > 0) {
      const timeoutToken = setTimeout((() => {
        response.emit("error", new Error("PAYLOAD_TIMEOUT"));
        request.destroy();
      }), timeoutPayload);
      response.on("end", (() => clearTimeout(timeoutToken)));
    }
    if (onProgressDownload) {
      const payloadLength = Number(response.headers["content-length"]) || 1 / 0;
      let chunkLength = 0;
      const onProgress = chunk => {
        chunkLength += chunk.length;
        onProgressDownload(chunkLength, payloadLength);
      };
      const {socket} = response;
      socket.on("data", onProgress);
      const clearListener = () => socket.off("data", onProgress);
      response.on("error", clearListener);
      response.on("end", clearListener);
    }
    return "gzip" === response.headers["content-encoding"] ? setupStreamPipe(response, createGunzip()) : response;
  };
  const _buffer = async () => readableStreamToBufferAsync(_stream());
  const _onReject = error => {
    throw tagError(error);
  };
  return {
    stream: () => {
      try {
        return _stream();
      } catch (error) {
        throw tagError(error);
      }
    },
    buffer: () => _buffer().catch(_onReject),
    arrayBuffer: () => _buffer().then(fromNodejsBuffer).catch(_onReject),
    text: () => _buffer().then((buffer => String(buffer))).catch(_onReject),
    json: () => _buffer().then((buffer => JSON.parse(String(buffer)))).catch(_onReject)
  };
};
const fetchWithJump = async (initialUrl, {fetch = fetchLikeRequest, jumpMax = 0, preFetch, ...option}) => {
  let url = initialUrl;
  let jumpCount = 0;
  let cookieList = [ option.headers && option.headers.cookie ].filter(Boolean);
  for (;;) {
    preFetch && await preFetch(url, jumpCount, cookieList);
    const response = await fetch(url, {
      ...option,
      headers: {
        ...option.headers,
        cookie: cookieList.join(";")
      }
    });
    if (!(response.status >= 300 && response.status <= 399 && response.headers.location)) return response;
    jumpCount++;
    if (jumpCount > jumpMax) throw new Error("JUMP_MAX_REACHED");
    url = new URL(response.headers.location, url).href;
    cookieList = [ ...cookieList, ...(response.headers["set-cookie"] || []).map((v => v.split(";")[0])) ];
  }
};
export { requestHttp, ping, fetchLikeRequest, fetchWithJump };