import { gzip, gzipSync, createGzip } from "node:zlib";
import { promisify } from "node:util";
import { DEFAULT_MIME, BASIC_EXTENSION_MAP } from "../../../common/module/MIME.js";
import { writeBufferToStreamAsync, quickRunletFromStream } from "../../data/Stream.js";
import { getEntityTagByContentHash } from "../../module/EntityTag.js";
const gzipAsync = promisify(gzip);
const checkPair = (store, entityTag, isEnableGzip = !1) => {
  entityTag && store.response.setHeader("etag", entityTag);
  const headers = store.request.headers;
  const shouldSend = !entityTag || !(headers["if-none-match"] || "").includes(entityTag);
  return [ shouldSend, isEnableGzip && shouldSend && (headers["accept-encoding"] || "").includes("gzip") ];
};
const responseHead = (store, type, length, [shouldSend, shouldGzip], range) => {
  let statusCode = 200;
  const headers = {
    "content-type": type
  };
  length >= 0 && (headers["content-length"] = length);
  if (range) {
    const [start, end] = range;
    statusCode = 206;
    headers["content-length"] = end - start + 1;
    headers["content-range"] = `bytes ${start}-${end}/${length}`;
    shouldSend = !0;
  } else shouldSend ? shouldGzip && (headers["content-encoding"] = "gzip") : statusCode = 304;
  store.response.writeHead(statusCode, headers);
  return shouldSend;
};
const responderSendBuffer = (store, {buffer, entityTag, type = DEFAULT_MIME, length = buffer.length}) => responseHead(store, type, length, checkPair(store, entityTag)) && writeBufferToStreamAsync(store.response, buffer);
const responderSendBufferRange = (store, {buffer, entityTag, type = DEFAULT_MIME, length = buffer.length}, [start, end]) => {
  end = Math.min(end, length - 1);
  return responseHead(store, type, length, checkPair(store, entityTag), [ start, end ]) && writeBufferToStreamAsync(store.response, buffer.slice(start, end + 1));
};
const responderSendBufferCompress = async (store, {buffer, bufferGzip, entityTag, type = DEFAULT_MIME, length = buffer.length}) => {
  const [shouldSend, shouldGzip] = checkPair(store, entityTag, !0);
  shouldGzip && (length = (buffer = bufferGzip || await gzipAsync(buffer)).length);
  return responseHead(store, type, length, [ shouldSend, shouldGzip ]) && writeBufferToStreamAsync(store.response, buffer);
};
const responderSendStream = (store, {stream, entityTag, type = DEFAULT_MIME, length}) => responseHead(store, type, length, checkPair(store, entityTag)) && quickRunletFromStream(stream, store.response);
const responderSendStreamRange = (store, {streamRange, entityTag, type = DEFAULT_MIME, length = "*"}, [start, end]) => {
  length >= 0 && (end = Math.min(end, length - 1));
  return responseHead(store, type, length, checkPair(store, entityTag), [ start, end ]) && quickRunletFromStream(streamRange, store.response);
};
const responderSendStreamCompress = async (store, {stream, streamGzip, entityTag, type = DEFAULT_MIME, length}) => {
  const [shouldSend, shouldGzip] = checkPair(store, entityTag, !0);
  return responseHead(store, type, shouldGzip ? void 0 : length, [ shouldSend, shouldGzip ]) && quickRunletFromStream.apply(null, shouldGzip ? streamGzip ? [ streamGzip, store.response ] : [ stream, createGzip(), store.response ] : [ stream, store.response ]);
};
const responderSendJSON = (store, {object, entityTag}) => responderSendBufferCompress(store, {
  buffer: Buffer.from(JSON.stringify(object)),
  type: BASIC_EXTENSION_MAP.json,
  entityTag
});
const prepareBufferData = (buffer, type) => ({
  buffer,
  type,
  bufferGzip: gzipSync(buffer),
  entityTag: getEntityTagByContentHash(buffer)
});
const createResponderFavicon = (buffer = Buffer.from("iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNMXvf/PwAGnQMR4CJUOAAAAABJRU5ErkJggg==", "base64"), type = BASIC_EXTENSION_MAP.png) => store => responderSendBufferCompress(store, prepareBufferData(buffer, type));
export { responderSendBuffer, responderSendBufferRange, responderSendBufferCompress, responderSendStream, responderSendStreamRange, responderSendStreamCompress, responderSendJSON, createResponderFavicon, prepareBufferData };