import { createReadStream } from "node:fs";
import { createCacheMap } from "../../../common/data/CacheMap.js";
import { getMIMETypeFromFileName } from "../../../common/module/MIME.js";
import { readBuffer } from "../../fs/File.js";
import { getPathStat } from "../../fs/Path.js";
import { getWeakEntityTagByStat } from "../../module/EntityTag.js";
import { responderEndWithStatusCode } from "./Common.js";
import { responderSendBuffer, responderSendBufferRange, responderSendStream, responderSendStreamRange } from "./Send.js";
const createDefaultCacheMap = () => createCacheMap({
  valueSizeSumMax: 33554432
});
const createResponderServeStatic = ({sizeSingleMax = 524288, expireTime = 6e4, isEnableGzip = !1, isEnableRange = !0, serveCacheMap = createDefaultCacheMap()}) => {
  const serveCache = async (store, filePath, filePathGz, rangePair) => {
    const filePathServe = filePathGz || filePath;
    const bufferData = serveCacheMap.get(filePathServe);
    if (!bufferData) return !1;
    filePathGz && store.response.setHeader("content-encoding", "gzip");
    isEnableRange && store.response.setHeader("accept-ranges", "bytes");
    rangePair ? await responderSendBufferRange(store, bufferData, rangePair) : await responderSendBuffer(store, bufferData);
    return !0;
  };
  const serve = async (store, filePath, filePathGz, rangePair) => {
    const filePathServe = filePathGz || filePath;
    const stat = await getPathStat(filePathServe);
    if (!stat.isFile()) return !1;
    const entityTag = getWeakEntityTagByStat(stat);
    const type = getMIMETypeFromFileName(filePath);
    const length = stat.size;
    filePathGz && store.response.setHeader("content-encoding", "gzip");
    isEnableRange && store.response.setHeader("accept-ranges", "bytes");
    if (rangePair) {
      rangePair[1] = Math.min(rangePair[1], length - 1);
      await responderSendStreamRange(store, {
        streamRange: createReadStream(filePathServe, {
          start: rangePair[0],
          end: rangePair[1]
        }),
        entityTag,
        type,
        length
      }, rangePair);
    } else if (length > sizeSingleMax) await responderSendStream(store, {
      stream: createReadStream(filePathServe),
      entityTag,
      type,
      length
    }); else {
      const bufferData = {
        buffer: await readBuffer(filePathServe),
        entityTag,
        type,
        length
      };
      serveCacheMap.set(filePathServe, bufferData, length, Date.now() + expireTime);
      await responderSendBuffer(store, bufferData);
    }
    return !0;
  };
  return async (store, filePath) => {
    if (await (async (store, filePath) => {
      if ("HEAD" !== store.request.method) return !1;
      let bufferData = serveCacheMap.get(filePath);
      if (!bufferData) {
        const stat = await getPathStat(filePath);
        if (!stat.isFile()) throw new Error(`miss file: ${filePath}`);
        bufferData = {
          entityTag: getWeakEntityTagByStat(stat),
          type: getMIMETypeFromFileName(filePath),
          length: stat.size
        };
      }
      const headerMap = {
        etag: bufferData.entityTag,
        "content-type": bufferData.type,
        "content-length": bufferData.length,
        "accept-ranges": isEnableRange ? "bytes" : "none"
      };
      await responderEndWithStatusCode(store, {
        statusCode: 200,
        headerMap
      });
      return !0;
    })(store, filePath)) return;
    const filePathGz = isEnableGzip && REGEXP_ENCODING_GZIP.test(store.request.headers["accept-encoding"]) && filePath + ".gz";
    const rangePair = isEnableRange && parseRangeHeader(store.request.headers.range);
    if (!(filePathGz && await serveCache(store, filePath, filePathGz) || await serveCache(store, filePath, void 0, rangePair) || filePathGz && await serve(store, filePath, filePathGz) || await serve(store, filePath, void 0, rangePair))) throw new Error(`miss file: ${filePath}`);
  };
};
const REGEXP_ENCODING_GZIP = /gzip/i;
const parseRangeHeader = rangeString => {
  const result = REGEXP_HEADER_RANGE.exec(rangeString);
  if (!result) return;
  const [, startString, endString] = result;
  const start = parseInt(startString);
  const end = endString ? parseInt(endString) : 1 / 0;
  return start < end ? [ start, end ] : void 0;
};
const REGEXP_HEADER_RANGE = /bytes=(\d+)-(\d+)?$/i;
export { createDefaultCacheMap, createResponderServeStatic };