import { networkInterfaces } from "node:os";
import { randomBytes } from "node:crypto";
import { createServer as createTCPServer } from "node:net";
import { createServer as createTLSServer } from "node:tls";
import { createServer as createHttpServer } from "node:http";
import { createServer as createHttpsServer } from "node:https";
import { clock } from "../../common/time.js";
import { isNumber, isBasicArray, isBasicObject } from "../../common/check.js";
import { prettyStringifyConfigObject } from "../../common/format.js";
import { createCacheMap2 } from "../../common/data/CacheMap2.js";
import { createStateStoreLite } from "../../common/immutable/StateStore.js";
import { objectFromEntries } from "../../common/immutable/Object.js";
import { responderError, responderEnd } from "./Responder/Common.js";
const createServerExot = ({id = "server", protocol, skipSessionPatch = !1, forceCloseTimeout = 2e3, ...option}) => {
  if (![ "tcp:", "tls:", "http:", "https:" ].includes(protocol)) throw new Error(`invalid protocol: ${protocol}`);
  const isSecure = [ "tls:", "https:" ].includes(protocol);
  (option = {
    isSecure,
    protocol,
    port: isSecure ? 443 : 80,
    hostname: "127.0.0.1",
    ...isSecure && !skipSessionPatch && {
      sessionTimeout: 14400
    },
    ...option
  }).baseUrl = `${protocol}//${option.hostname}:${option.port}`
  /** @type { ServerSecure | undefined } */;
  const serverSec = isSecure ? "tls:" === protocol ? createTLSServer(option) : createHttpsServer(option) : void 0;
  /** @type { ServerBase } */  const server = serverSec || ("tcp:" === protocol ? createTCPServer() : createHttpServer());
  serverSec && !skipSessionPatch && (server => {
    const sessionCacheMap = createCacheMap2({
      sizeMax: 4096,
      expireAfter: 6e5
    });
    server.on("newSession", ((sessionId, sessionData, next) => {
      sessionCacheMap.set(sessionId.toString("base64"), sessionData);
      next();
    }));
    server.on("resumeSession", ((sessionId, next) => {
      next(null, sessionCacheMap.get(sessionId.toString("base64")) || null);
    }));
  })(serverSec);
  const socketSet = new Set;
  server.on("connection", (socket => {
    socketSet.add(socket);
    socket.once("close", (() => {
      socketSet.delete(socket);
    }));
  }));
  let sessionTicketRotateToken;
  return {
    id,
    up: async () => !server.listening && new Promise(((resolve, reject) => {
      server.on("error", reject);
      server.listen(option.port, option.hostname.replace(/[[\]]/g, ""), (() => {
        server.off("error", reject);
        if (serverSec && !skipSessionPatch) {
          const resetSessionTicketKey = () => serverSec.setTicketKeys(randomBytes(48));
          sessionTicketRotateToken && clearInterval(sessionTicketRotateToken);
          sessionTicketRotateToken = setInterval(resetSessionTicketKey, 144e5);
          resetSessionTicketKey();
        }
        resolve();
      }));
    })),
    down: async () => server.listening && new Promise((resolve => {
      sessionTicketRotateToken && clearInterval(sessionTicketRotateToken);
      const forceCloseToken = isNumber(forceCloseTimeout) && forceCloseTimeout !== 1 / 0 && setTimeout((() => {
        for (const socket of socketSet) socket.destroy();
      }), forceCloseTimeout);
      server.close((() => {
        forceCloseToken && clearTimeout(forceCloseToken);
        resolve();
      }));
    })),
    isUp: () => server.listening,
    server,
    option,
    socketSet
  };
};
const DEFAULT_RESPONDER_ERROR = responderError;
const DEFAULT_RESPONDER_END = responderEnd;
/** @typedef { Error & { status: number } } ErrorWithStatus */
/** @typedef { { time: number, status: number, error: ErrorWithStatus | null } } ConnState */
/** @typedef { { getState: () => ConnState, setState: (nextState: Partial<ConnState>) => ConnState, socket: Socket, request: IncomingMessage, response: ServerResponse } } ConnStore */
/** @typedef { (store: ConnStore, ...ext: any[]) => any | Promise<any> } Responder */
/** @type { (opt: { responderList: Responder[], responderError?: Responder, responderEnd?: Responder }) => (request: IncomingMessage, response: ServerResponse) => Promise<void> } */ const createRequestListener = ({responderList = [], responderError = DEFAULT_RESPONDER_ERROR, responderEnd = DEFAULT_RESPONDER_END}) => async (request, response) => {
  const store = /** @type { ConnStore } */ /** @type { unknown } */ createStateStoreLite({
    time: clock(),
    status: 500,
    error: null
  });
  store.socket = request.socket;
  store.request = request;
  store.response = response;
  try {
    for (const responder of responderList) await responder(store);
  } catch (error) {
    await responderError(store, error);
  }
  await responderEnd(store);
};
const describeServerOption = ({baseUrl, protocol, hostname, port}, title = "server", extraOption) => `[${title}]\n${prettyStringifyConfigObject({
  pid: process.pid,
  baseUrl,
  ...hostname && "0.0.0.0" !== hostname && "[::]" !== hostname ? {} : objectFromEntries(Object.entries(networkInterfaces()).reduce(((o, [interfaceName, interfaceList]) => {
    interfaceList.forEach(((/** @type { (typeof interfaceList)[0] & { isIPv4?: boolean } } */ {address, family, isIPv4 = "IPv4" === family}) => (hostname.startsWith("[") || isIPv4) && o.push([ isIPv4 ? address : `[${address}]`, interfaceName ])));
    return o;
  }), []).map((([address, interfaceName]) => [ `localUrl[${interfaceName}]`, `${protocol}//${address}:${port}` ]))),
  ...isBasicArray(extraOption) && extraOption.length ? {
    extraOption
  } : isBasicObject(extraOption) ? extraOption : {}
}, "  ", "  ")}`;
export { createServerExot, createRequestListener, describeServerOption };