import { arch, platform, release, hostname, cpus, networkInterfaces, totalmem, freemem, loadavg, uptime } from "node:os";
import { getHeapStatistics } from "node:v8";
import { percent, time, binary } from "../../common/format.js";
import { indentLine, indentList } from "../../common/string.js";
import { readTextSync } from "../fs/File.js";
const getSystemPlatform = () => ({
  platform: platform(),
  arch: arch(),
  release: release()
});
const describeSystemPlatform = ({platform, arch, release} = getSystemPlatform()) => `${platform} ${arch} [${release}]`;
const getSystemProcessor = () => cpus() || [ {
  model: "N/A",
  speed: "N/A",
  times: {
    user: 0,
    nice: 0,
    sys: 0,
    idle: 0,
    irq: 0
  }
} ];
const describeSystemProcessor = (processorList = getSystemProcessor()) => processorList.map((({model, speed, times}) => `[${model}] speed:${speed}MHz ${Object.entries(times).map((([k, v]) => `${k}:${time(v)}`)).join(" ")}`)).join("\n");
const getSystemMemory = () => {
  try {
    if ("linux" === platform()) {
      const stringMemInfo = readTextSync("/proc/meminfo");
      const toByte = regexp => 1024 * parseInt(regexp.exec(stringMemInfo)[1]);
      return {
        total: toByte(/MemTotal:\s+(\d+)/),
        free: toByte(/MemAvailable:\s+(\d+)/),
        swapTotal: toByte(/SwapTotal:\s+(\d+)/),
        swapFree: toByte(/SwapFree:\s+(\d+)/)
      };
    }
  } catch (error) {}
  return {
    total: totalmem(),
    free: freemem(),
    swapTotal: 0,
    swapFree: 0
  };
};
const describeSystemMemory = ({total, free, swapTotal, swapFree} = getSystemMemory()) => [ `Used: ${percent((total - free) / total)}${swapTotal ? `+${percent((swapTotal - swapFree) / swapTotal)}` : ""}`, `Total: ${binary(total)}B${swapTotal ? `+${binary(swapTotal)}B` : ""}`, `Free: ${binary(free)}B${swapTotal ? `+${binary(swapFree)}B` : ""}` ].join("\n");
const getSystemNetwork = () => ({
  hostname: hostname(),
  networkInterface: networkInterfaces()
});
const describeSystemNetwork = ({hostname, networkInterface} = getSystemNetwork()) => [ `[hostname] ${hostname}`, ...Object.entries(networkInterface).map((([name, addressList]) => indentList(`[interface] ${name}`, addressList.map((({address, netmask, mac, internal, family, cidr}) => `[${family}${internal ? "|INTERNAL" : ""}] ${cidr || address} (${mac})`))))) ].join("\n");
const getSystemActivity = () => ({
  uptime: 1e3 * uptime(),
  loadAverageList: loadavg()
});
const describeSystemActivity = ({uptime, loadAverageList} = getSystemActivity()) => [ `[uptime] ${time(uptime)}`, "win32" !== platform() && `[load average] ${loadAverageList.map(percent).join(", ")} (1min, 5min, 15min)` ].filter(Boolean).join("\n");
const getSystemStatus = () => ({
  platform: getSystemPlatform(),
  processor: getSystemProcessor(),
  memory: getSystemMemory(),
  network: getSystemNetwork(),
  activity: getSystemActivity()
});
const describeSystemStatus = ({platform, processor, memory, network, activity} = getSystemStatus()) => Object.entries({
  platform: describeSystemPlatform(platform),
  processor: describeSystemProcessor(processor),
  memory: describeSystemMemory(memory),
  network: describeSystemNetwork(network),
  activity: describeSystemActivity(activity)
}).map((([k, v]) => `[${k}]\n${indentLine(v)}`)).join("\n");
const getSystemInfo = () => {
  const toGiB = byte => Math.round(byte / 2 ** 30);
  const toPercent = (total, free) => total ? `${(100 * (1 - free / total)).toFixed(0)}%` : "-";
  const {total, free, swapTotal, swapFree} = getSystemMemory();
  const nCPU = cpus().length;
  const nMEM = toGiB(total);
  const nSWAP = toGiB(swapTotal);
  const loadAvgList = loadavg();
  const systemStatus = [ swapTotal ? `[${nCPU}CPU|${nMEM}+${nSWAP}G]` : `[${nCPU}CPU|${nMEM}G]`, swapTotal ? `mem/swap=${toPercent(total, free)}/${toPercent(swapTotal, swapFree)}` : `mem=${toPercent(total, free)}`, `loadavg=${loadAvgList.map((v => v.toFixed(1))).join("|")}`, `uptime=${time(1e3 * uptime())}` ].join(" ");
  const [load1Min, load5Min, load15Min] = loadAvgList;
  return {
    nCPU,
    nMEM,
    nSWAP,
    systemStatus,
    loadStatus: load15Min > 1.1 * nCPU && load5Min > .7 * nCPU && load1Min > .7 * nCPU ? "15MIN" : load5Min > 1.3 * nCPU && load1Min > .7 * nCPU ? "5MIN" : load5Min > .7 * nCPU && load1Min > 1.7 * nCPU ? "1MIN" : ""
  };
};
const V8_HEAP_RESERVED_SIZE = 1024 * (totalmem() < 2147483648 ? 24 : 48) * 1024;
const getV8HeapStatus = () => {
  const {used_heap_size: v8HeapUsed, heap_size_limit: v8HeapMax} = getHeapStatistics();
  const total = v8HeapMax - V8_HEAP_RESERVED_SIZE;
  return {
    total,
    free: Math.max(total - v8HeapUsed, 0)
  };
};
export { getSystemPlatform, getSystemProcessor, getSystemMemory, getSystemNetwork, getSystemActivity, getSystemStatus, describeSystemPlatform, describeSystemProcessor, describeSystemMemory, describeSystemNetwork, describeSystemActivity, describeSystemStatus, getSystemInfo, V8_HEAP_RESERVED_SIZE, getV8HeapStatus };